"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, TypedDict

from typing_extensions import NotRequired


if TYPE_CHECKING:
    from .common import LocalizedString, LanguageCode
    from .relationship import RelationshipResponse
    from .tags import TagResponse

from ..enums import ContentRating as _ContentRating
from ..enums import MangaRelationType as _MangaRelationType
from ..enums import MangaState as _MangaState
from ..enums import MangaStatus as _MangaStatus
from ..enums import PublicationDemographic as _PublicationDemographic
from ..enums import ReadingStatus as _ReadingStatus


__all__ = (
    "MangaLinks",
    "MangaAttributesResponse",
    "MangaRelationAttributesResponse",
    "MangaResponse",
    "GetMangaResponse",
    "MangaRelationCreateResponse",
    "MangaSearchResponse",
    "MangaRelation",
    "MangaRelationResponse",
    "ChaptersResponse",
    "VolumesAndChaptersResponse",
    "GetMangaVolumesAndChaptersResponse",
    "MangaReadMarkersResponse",
    "MangaGroupedReadMarkersResponse",
    "MangaSingleReadingStatusResponse",
    "MangaMultipleReadingStatusResponse",
)


class MangaLinks(TypedDict, total=False):
    """
    Please see here for more explicit info on these:-
        https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data


    al: Optional[:class:`str`]
        anilist key

    ap: Optional[:class:`str`]
        animeplanet key

    bw: Optional[:class:`str`]
        bookwalker.jp key

    mu: Optional[:class:`str`]
        mangaupdates key

    nu: Optional[:class:`str`]
        novelupdates key

    kt: Optional[:class:`str`]
        kitsu.io key

    amz: Optional[:class:`str`]
        amazon key

    ebj: Optional[:class:`str`]
        ebookjapan key

    mal: Optional[:class:`str`]
        myanimelist key

    cdj: Optional[:class:`str`]
        cdjapan key

    raw: Optional[:class:`str`]

    engtl: Optional[:class:`str`]
    """

    al: Optional[str]
    ap: Optional[str]
    bw: Optional[str]
    mu: Optional[str]
    nu: Optional[str]
    kt: Optional[str]
    amz: Optional[str]
    ebj: Optional[str]
    mal: Optional[str]
    cdj: Optional[str]
    raw: Optional[str]
    engtl: Optional[str]


class MangaAttributesResponse(TypedDict):
    """
    title: :class:`~hondana.types.common.LocalizedString`

    altTitle: List[:class:`~hondana.types.common.LocalizedString`]

    description: :class:`~hondana.types.common.LocalizedString`

    isLocked: :class:`bool`

    links: :class:`~hondana.types.manga.MangaLinks`

    originalLanguage: :class:`str`

    lastVolume: Optional[:class:`str`]

    lastChapter: Optional[:class:`str`]

    publicationDemographic: Optional[:class:`~hondana.PublicationDemographic`]

    status: :class:`~hondana.MangaStatus`

    year: Optional[:class:`int`]

    contentRating: Optional[:class:`~hondana.ContentRating`]

    chapterNumbersResetOnNewVolume: :class:`bool`

    latestUploadedChapter: :class:`str`

    availableTranslatedLanguages: List[:class:`~hondana.types.common.LanguageCode`]

    tags: List[:class:`~hondana.types.tags.TagResponse`]

    state: :class:`~hondana.MangaState`

    version: :class:`int`

    createdAt: :class:`str`

    updatedAt: :class:`str`

    isLocked: Optional[:class:`bool`]
        This key is optional
    """

    title: LocalizedString
    altTitles: list[LocalizedString]
    description: LocalizedString
    isLocked: bool
    links: MangaLinks
    originalLanguage: str
    lastVolume: Optional[str]
    lastChapter: Optional[str]
    publicationDemographic: Optional[_PublicationDemographic]
    status: _MangaStatus
    year: Optional[int]
    contentRating: Optional[_ContentRating]
    chapterNumbersResetOnNewVolume: bool
    latestUploadedChapter: str
    availableTranslatedLanguages: list[LanguageCode]
    tags: list[TagResponse]
    state: _MangaState
    version: int
    createdAt: str
    updatedAt: str


class MangaRelationAttributesResponse(TypedDict):
    """
    relation: :class:`~hondana.MangaRelationType`

    version: int
    """

    relation: _MangaRelationType
    version: int


class MangaResponse(TypedDict):
    """
    id: :class:`str`

    type: Literal[``"manga"``]

    attributes: :class:`~hondana.types.manga.MangaAttributesResponse`

    relationships: List[:class:`~hondana.types.relationship.RelationshipResponse`]
        This key is optional, in the event this payload is gotten from the "relationships" of another object.

        This key can contain minimal or full data depending on the ``includes[]`` parameter of its request.
        See here for more info: https://api.mangadex.org/docs.html#section/Reference-Expansion

    related: :class:`~hondana.MangaRelationType`
        This key is optional.
    """

    id: str
    type: Literal["manga"]
    attributes: MangaAttributesResponse
    relationships: NotRequired[list[RelationshipResponse]]
    related: NotRequired[_MangaRelationType]


class GetMangaResponse(TypedDict):
    """
    result: Literal[``"ok"``, ``"error"``]

    response: Literal[``"entity"``]

    data: :class:`~hondana.types.manga.MangaResponse`
    """

    result: Literal["ok", "error"]
    response: Literal["entity"]
    data: MangaResponse


class MangaRelationCreateResponse(TypedDict):
    """
    result: Literal[``"ok"``, ``"error"``]

    response: Literal[``"entity"``]

    data: :class:`~hondana.types.manga.MangaRelation`
    """

    result: Literal["ok", "error"]
    response: Literal["entity"]
    data: MangaRelation


class MangaSearchResponse(TypedDict):
    """
    result: Literal[``"ok"``, ``"error"``]

    response: Literal[``"collection"``]

    data: List[:class:`~hondana.types.manga.MangaResponse`]

    limit: :class:`int`

    offset: :class:`int`

    total: :class:`int`
    """

    result: Literal["ok", "error"]
    response: Literal["collection"]
    data: list[MangaResponse]
    limit: int
    offset: int
    total: int


class MangaRelation(TypedDict):
    """
    id: :class:`str`

    type: Literal[``"manga_relation"``]

    attributes: :class:`~hondana.types.manga.MangaRelationAttributesResponse`

    relationships: List[:class:`~hondana.types.relationship.RelationshipResponse`]
        This key is optional, in the event this payload is gotten from the "relationships" of another object.

        This key can contain minimal or full data depending on the ``includes[]`` parameter of its request.
        See here for more info: https://api.mangadex.org/docs.html#section/Reference-Expansion

    related: :class:`~hondana.MangaRelationType`
        This key is optional.
    """

    id: str
    type: Literal["manga_relation"]
    attributes: MangaRelationAttributesResponse
    relationships: NotRequired[list[RelationshipResponse]]
    related: NotRequired[_MangaRelationType]


class MangaRelationResponse(TypedDict):
    """
    result: Literal[``"ok"``, ``"error"``]

    response: Literal[``"collection"``]

    data: List[:class:`~hondana.types.manga.MangaRelation`]

    limit: :class:`int`

    offset: :class:`int`

    total: :class:`int`
    """

    result: Literal["ok", "error"]
    response: Literal["collection"]
    data: list[MangaRelation]
    limit: int
    offset: int
    total: int


class ChaptersResponse(TypedDict):
    """
    chapter: :class:`str`

    count: :class:`str`
    """

    chapter: str
    count: str


class VolumesAndChaptersResponse(TypedDict, total=False):
    """
    chapters: Dict[:class:`str`, :class:`~hondana.types.manga.ChaptersResponse`]

    count: :class:`int`

    volume: :class:`str`
    """

    chapters: dict[str, ChaptersResponse]
    count: int
    volume: str


class GetMangaVolumesAndChaptersResponse(TypedDict):
    """
    result: Literal[``"ok"``, ``"error"``]

    volumes: Optional[Dict[:class:`str`, :class:`~hondana.types.manga.VolumesAndChaptersResponse`]]
    """

    result: Literal["ok", "error"]
    volumes: Optional[dict[str, VolumesAndChaptersResponse]]


class MangaReadMarkersResponse(TypedDict):
    """
    result: Literal[``"ok"``]

    data: List[:class:`str`]
    """

    result: Literal["ok"]
    data: list[str]


class MangaGroupedReadMarkersResponse(TypedDict):
    """
    result: Literal[``"ok"``]

    data: Dict[:class:`str`, List[:class:`str`]]
    """

    result: Literal["ok"]
    data: dict[str, list[str]]


class MangaSingleReadingStatusResponse(TypedDict):
    """
    result: Literal[``"ok"``]

    status: :class:`~hondana.ReadingStatus`
    """

    result: Literal["ok"]
    status: _ReadingStatus


class MangaMultipleReadingStatusResponse(TypedDict):
    """
    result: Literal[``"ok"``]

    statuses: Dict[str, :class:`~hondana.ReadingStatus`]
        Mapping of [``manga_id``: :class:`~hondana.ReadingStatus` value]
    """

    result: Literal["ok"]
    statuses: dict[str, _ReadingStatus]



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .enums import CustomListVisibility
from .manga import Manga
from .query import MangaIncludes
from .user import User
from .utils import relationship_finder, require_authentication


if TYPE_CHECKING:
    from .http import HTTPClient
    from .types.custom_list import CustomListResponse
    from .types.manga import MangaResponse
    from .types.relationship import RelationshipResponse
    from .types.user import UserResponse


__all__ = ("CustomList",)


class CustomList:
    """A class representing a CustomList returned from the MangaDex API.

    Attributes
    -----------
    id: :class:`str`
        The UUID relating to this custom list.
    name: :class:`str`
        The name of this custom list.
    visibility: :class:`~hondana.CustomListVisibility`
        The visibility of this custom list.
    version: :class:`int`
        The version revision of this custom list.
    """

    __slots__ = (
        "_http",
        "_data",
        "_attributes",
        "id",
        "name",
        "visibility",
        "version",
        "_owner_relationship",
        "_manga_relationships",
        "__owner",
        "__manga",
    )

    def __init__(self, http: HTTPClient, payload: CustomListResponse) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        relationships: list[RelationshipResponse] = self._data.pop("relationships", [])  # type: ignore # can't pop from a TypedDict
        self.id: str = self._data["id"]
        self.name: str = self._attributes["name"]
        self.visibility: CustomListVisibility = CustomListVisibility(self._attributes["visibility"])
        self.version: int = self._attributes["version"]
        self._owner_relationship: Optional[UserResponse] = relationship_finder(relationships, "user", limit=1)
        self._manga_relationships: list[MangaResponse] = relationship_finder(relationships, "manga", limit=None)
        self.__owner: Optional[User] = None
        self.__manga: Optional[list[Manga]] = None

    def __repr__(self) -> str:
        return f"<CustomList id={self.id!r} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: CustomList) -> bool:
        return isinstance(other, CustomList) and self.id == other.id

    def __ne__(self, other: CustomList) -> bool:
        return not self.__eq__(other)

    @property
    def url(self) -> str:
        """The URL to this custom list.

        Returns
        --------
        :class:`str`
            The URL of the custom list.
        """
        return f"https://mangadex.org/list/{self.id}"

    @property
    def owner(self) -> Optional[User]:
        """Returns the owner of this custom list.

        Returns
        --------
        Optional[:class:`~hondana.User`]
            The owner of this list, if any.
        """
        if self.__owner is not None:
            return self.__owner

        if not self._owner_relationship:
            return None

        if "attributes" in self._owner_relationship:
            self.__owner = User(self._http, self._owner_relationship)
            return self.__owner

    @owner.setter
    def owner(self, other: User) -> None:
        if isinstance(other, User):
            self.__owner = other

    @property
    def manga(self) -> Optional[list[Manga]]:
        """Returns the of manga present in this custom list, if any.

        Returns
        --------
        Optional[List[:class:`~hondana.Manga`]]
            The list of manga present, if any.
        """

        if self.__manga is not None:
            return self.__manga

        if not self._manga_relationships:
            return

        fmt: list[Manga] = [Manga(self._http, manga) for manga in self._manga_relationships if "attributes" in manga]
        if not fmt:
            return

        return fmt

    @manga.setter
    def manga(self, other: list[Manga]) -> None:
        fmt = [item for item in other if isinstance(item, Manga)]
        self.__manga = fmt

    async def get_owner(self) -> Optional[User]:
        """|coro|

        This method will make an API request to get the owner of a Custom List.

        Returns
        --------
        Optional[:class:`~hondana.User`]
            The owner of this list, if any.
        """
        if self.owner is not None:
            return self.owner

        if not self._owner_relationship:
            return

        data = await self._http._get_user(self._owner_relationship["id"])
        self.__owner = User(self._http, data["data"])
        return self.__owner

    async def get_manga(self) -> Optional[list[Manga]]:
        """|coro|"""
        if self.manga is not None:
            return self.manga

        if not self._manga_relationships:
            return

        ids = [r["id"] for r in self._manga_relationships]

        data = await self._http._manga_list(
            limit=100,
            offset=0,
            title=None,
            authors=None,
            artists=None,
            year=None,
            included_tags=None,
            excluded_tags=None,
            status=None,
            original_language=None,
            excluded_original_language=None,
            available_translated_language=None,
            publication_demographic=None,
            ids=ids,
            content_rating=None,
            created_at_since=None,
            updated_at_since=None,
            order=None,
            includes=MangaIncludes(),
            has_available_chapters=None,
            group=None,
        )

        ret: list[Manga] = [Manga(self._http, item) for item in data["data"]]
        if not ret:
            return

        return ret

    @require_authentication
    async def update(
        self,
        *,
        name: Optional[str] = None,
        visibility: Optional[CustomListVisibility] = None,
        manga: Optional[list[str]] = None,
        version: int,
    ) -> CustomList:
        """|coro|

        This method will update the current custom list within the MangaDex API.

        Parameters
        -----------
        name: Optional[:class:`str`]
            The name we wish to edit the custom list with.
        visibility: Optional[:class:`~hondana.CustomListVisibility`]
            The visibility we wish to edit the custom list with.
        manga: Optional[List[:class:`str`]]
            The list of manga IDs to edit this custom list with.
        version: :class:`int`
            The version revision of this custom list.


        .. note::
            Updating a custom list is an atomic action.
            Passing the ``manga`` key here will overwrite the manga in this custom list.

        Raises
        -------
        :exc:`BadRequest`
            The request body was malformed.
        :exc:`Forbidden`
            You are not authorized to edit this custom list.
        :exc:`NotFound`
            The custom list was not found, or one of the manga passed was not found.

        Returns
        --------
        :class:`CustomList`
            The returned custom list after it was updated.
        """
        data = await self._http._update_custom_list(self.id, name=name, visibility=visibility, manga=manga, version=version)

        return self.__class__(self._http, data["data"])

    @require_authentication
    async def delete_custom_list(self) -> None:
        """|coro|

        This method will delete the current custom list from the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            You are not authorized to delete this custom list.
        :exc:`NotFound`
            The custom list with this UUID was not found.
        """
        await self._http._delete_custom_list(self.id)

    @require_authentication
    async def follow(self) -> None:
        """|coro|

        The method will follow a custom list within the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            You are not authorized to follow this custom list.
        :exc:`NotFound`
            The specified custom list does not exist.
        """
        await self._http._follow_custom_list(self.id)

    @require_authentication
    async def unfollow(self) -> None:
        """|coro|

        The method will unfollow a custom list within the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            You are not authorized to unfollow this custom list.
        :exc:`NotFound`
            The specified custom list does not exist.
        """
        await self._http._unfollow_custom_list(self.id)



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar


if TYPE_CHECKING:
    from .author import Author
    from .chapter import Chapter, PreviouslyReadChapter
    from .cover import Cover
    from .custom_list import CustomList
    from .http import HTTPClient
    from .legacy import LegacyItem
    from .manga import Manga, MangaRelation
    from .report import Report, UserReport
    from .scanlator_group import ScanlatorGroup
    from .types.author import GetMultiAuthorResponse
    from .types.chapter import ChapterReadHistoryResponse, GetMultiChapterResponse
    from .types.cover import GetMultiCoverResponse
    from .types.custom_list import GetMultiCustomListResponse
    from .types.legacy import GetLegacyMappingResponse
    from .types.manga import MangaRelationResponse, MangaSearchResponse
    from .types.report import GetReportReasonResponse, GetUserReportReasonResponse
    from .types.scanlator_group import GetMultiScanlationGroupResponse
    from .types.user import GetMultiUserResponse
    from .user import User

__all__ = (
    "MangaCollection",
    "MangaRelationCollection",
    "ChapterFeed",
    "AuthorCollection",
    "CoverCollection",
    "ScanlatorGroupCollection",
    "ReportCollection",
    "UserReportCollection",
    "UserCollection",
    "CustomListCollection",
    "LegacyMappingCollection",
    "ChapterReadHistoryCollection",
)

T = TypeVar("T")


class BaseCollection(ABC, Generic[T]):
    """
    The base class for all collections. This class serves to make it easier to create functions that process
    arbitrary collections without needing to set up ``isinstance()`` checks.

    Attributes
    -----------
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    total: int
    offset: int
    limit: int

    @property
    @abstractmethod
    def items(self) -> list[T]:
        """
        Returns the items in the collection.

        Returns
        -------

        :class:`list`
        """


class MangaCollection(BaseCollection["Manga"]):
    """
    A collection object type to represent manga.

    Attributes
    -----------
    manga: List[:class:`~hondana.Manga`]
        The manga returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "manga",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: MangaSearchResponse, manga: list[Manga]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: MangaSearchResponse = payload
        self.manga: list[Manga] = manga
        self.total: int = payload.get("total", 0)
        self.offset: int = payload.get("offset", 0)
        self.limit: int = payload.get("limit", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<MangaFeed manga={len(self.manga)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[Manga]:
        """
        Returns the mangas in the collection.

        Returns
        -------

        List[:class:`~hondana.Manga`]
        """
        return self.manga


class MangaRelationCollection(BaseCollection["MangaRelation"]):
    """
    A collection object type to represent manga relations.

    Attributes
    -----------
    relations: List[:class:`~hondana.MangaRelation`]
        The manga relations returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "relations",
        "total",
        "offset",
        "limit",
    )

    def __init__(self, http: HTTPClient, payload: MangaRelationResponse, relations: list[MangaRelation]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data = payload
        self.relations: list[MangaRelation] = relations
        self.total: int = payload.get("total", 0)
        self.offset: int = payload.get("offset", 0)
        self.limit: int = payload.get("limit", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<MangaRelationCollection authors={len(self.relations)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[MangaRelation]:
        """
        Returns the manga relations in the collection.

        Returns
        -------

        List[:class:`~hondana.MangaRelation`]
        """
        return self.relations


class ChapterFeed(BaseCollection["Chapter"]):
    """
    A collection object type to represent chapters.

    Attributes
    -----------
    chapters: List[:class:`~hondana.Chapter`]
        The chapters returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "chapters",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiChapterResponse, chapters: list[Chapter]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiChapterResponse = payload
        self.chapters: list[Chapter] = chapters
        self.total: int = payload.get("total", 0)
        self.offset: int = payload.get("offset", 0)
        self.limit: int = payload.get("limit", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<ChapterFeed chapters={len(self.chapters)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[Chapter]:
        """
        Returns the chapters in the collection.

        Returns
        -------

        List[:class:`~hondana.Chapter`]
        """
        return self.chapters


class AuthorCollection(BaseCollection["Author"]):
    """
    A collection object type to represent authors.

    Attributes
    -----------
    authors: List[:class:`~hondana.Author`]
        The authors returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "authors",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiAuthorResponse, authors: list[Author]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiAuthorResponse = payload
        self.authors: list[Author] = authors
        self.total: int = payload.get("total", 0)
        self.offset: int = payload.get("offset", 0)
        self.limit: int = payload.get("limit", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<ArtistCollection authors={len(self.authors)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[Author]:
        """
        Returns the authors in the collection.

        Returns
        -------

        List[:class:`~hondana.Author`]
        """
        return self.authors


class CoverCollection(BaseCollection["Cover"]):
    """
    A collection object type to represent covers.

    Attributes
    -----------
    covers: List[:class:`~hondana.Cover`]
        The covers returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "covers",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiCoverResponse, covers: list[Cover]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiCoverResponse = payload
        self.covers: list[Cover] = covers
        self.total: int = payload.get("total", 0)
        self.offset: int = payload.get("offset", 0)
        self.limit: int = payload.get("limit", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<CoverCollection covers={len(self.covers)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[Cover]:
        """
        Returns the covers in the collection.

        Returns
        -------

        List[:class:`~hondana.Cover`]
        """
        return self.covers


class ScanlatorGroupCollection(BaseCollection["ScanlatorGroup"]):
    """
    A collection object type to represent scanlator groups.

    Attributes
    -----------
    groups: List[:class:`~hondana.ScanlatorGroup`]
        The groups returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "groups",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiScanlationGroupResponse, groups: list[ScanlatorGroup]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiScanlationGroupResponse = payload
        self.groups: list[ScanlatorGroup] = groups
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<ScanlatorGroupCollection groups={len(self.groups)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[ScanlatorGroup]:
        """
        Returns the groups in the collection.

        Returns
        -------

        List[:class:`~hondana.ScanlatorGroup`]
        """
        return self.groups


class ReportCollection(BaseCollection["Report"]):
    """
    A collection object type to represent reports.

    Attributes
    -----------
    reports: List[:class:`~hondana.Report`]
        The reports returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "reports",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetReportReasonResponse, reports: list[Report]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetReportReasonResponse = payload
        self.reports: list[Report] = reports
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<ReportCollection reports={len(self.reports)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[Report]:
        """
        Returns the reports in the collection.

        Returns
        -------

        List[:class:`~hondana.Report`]
        """
        return self.reports


class UserReportCollection(BaseCollection["UserReport"]):
    """
    A collection object type to represent reports.

    Attributes
    -----------
    reports: List[:class:`~hondana.UserReport`]
        The reports returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "reports",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetUserReportReasonResponse, reports: list[UserReport]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetUserReportReasonResponse = payload
        self.reports: list[UserReport] = reports
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return (
            f"<UserReportCollection reports={len(self.reports)} total={self.total} offset={self.offset} limit={self.limit}>"
        )

    @property
    def items(self) -> list[UserReport]:
        """
        Returns the reports in the collection.

        Returns
        -------

        List[:class:`~hondana.UserReport`]
        """
        return self.reports


class UserCollection(BaseCollection["User"]):
    """
    A collection object type to represent users.

    Attributes
    -----------
    users: List[:class:`~hondana.User`]
        The users returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "users",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiUserResponse, users: list[User]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiUserResponse = payload
        self.users: list[User] = users
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<UserCollection users={len(self.users)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[User]:
        """
        Returns the users in the collection.

        Returns
        -------

        List[:class:`~hondana.User`]
        """
        return self.users


class CustomListCollection(BaseCollection["CustomList"]):
    """
    A collection object type to represent custom lists.

    Attributes
    -----------
    lists: List[:class:`~hondana.CustomList`]
        The custom lists returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "lists",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetMultiCustomListResponse, lists: list[CustomList]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetMultiCustomListResponse = payload
        self.lists: list[CustomList] = lists
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<CustomListCollection lists={len(self.lists)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[CustomList]:
        """
        Returns the custom lists in the collection.

        Returns
        -------

        List[:class:`~hondana.CustomList`]
        """
        return self.lists


class LegacyMappingCollection(BaseCollection["LegacyItem"]):
    """
    A collection object type to represent custom lists.

    Attributes
    -----------
    legacy_mappings: List[:class:`~hondana.LegacyItem`]
        The legacy mappings returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "legacy_mappings",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: GetLegacyMappingResponse, mappings: list[LegacyItem]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data: GetLegacyMappingResponse = payload
        self.legacy_mappings: list[LegacyItem] = mappings
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<LegacyMappingCollection legacy_mappings={len(self.legacy_mappings)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[LegacyItem]:
        """
        Returns the legacy mappings in the collection.

        Returns
        -------

        List[:class:`~hondana.LegacyItem`]
        """
        return self.legacy_mappings


class ChapterReadHistoryCollection(BaseCollection["PreviouslyReadChapter"]):
    """
    A collection object type to represent chapter read history.

    Attributes
    -----------
    chapter_read_histories: List[:class:`~hondana.PreviouslyReadChapter`]
        The chapter read histories returned from this collection.
    total: :class:`int`
        The total possible results with this query could return.
    offset: :class:`int`
        The offset used in this query.
    limit: :class:`int`
        The limit used in this query.
    """

    __slots__ = (
        "_http",
        "_data",
        "chapter_read_histories",
        "total",
        "limit",
        "offset",
    )

    def __init__(self, http: HTTPClient, payload: ChapterReadHistoryResponse, history: list[PreviouslyReadChapter]) -> None:
        self._http: HTTPClient = http
        payload.pop("data", [])  # type: ignore # can't pop from a TypedDict
        self._data = payload
        self.history: list[PreviouslyReadChapter] = history
        self.total: int = payload.get("total", 0)
        self.limit: int = payload.get("limit", 0)
        self.offset: int = payload.get("offset", 0)
        super().__init__()

    def __repr__(self) -> str:
        return f"<ChapterReadHistoryCollection history={len(self.history)} total={self.total} offset={self.offset} limit={self.limit}>"

    @property
    def items(self) -> list[PreviouslyReadChapter]:
        """
        Returns the legacy mappings in the collection.

        Returns
        -------

        List[:class:`~hondana.PreviouslyReadChapter`]
        """
        return self.history



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .query import ScanlatorGroupIncludes
from .scanlator_group import ScanlatorGroup
from .utils import relationship_finder, require_authentication


if TYPE_CHECKING:
    from .http import HTTPClient
    from .types.relationship import RelationshipResponse
    from .types.scanlator_group import ScanlationGroupResponse
    from .types.user import UserResponse

__all__ = ("User",)


class User:
    """
    A class representing a user from the MangaDex API.

    Attributes
    -----------
    id: :class:`str`
        The user's UUID.
    username: :class:`str`
        The user's username.
    version: :class:`int`
        The user's version revision.
    roles: List[:class:`str`]
        The list of roles this person has.


    .. note::
        Unlike most other api objects, this type does not have related relationship properties due to not returning a full ``relationships`` key.


    """

    __slots__ = (
        "_http",
        "_data",
        "_attributes",
        "_relationships",
        "id",
        "username",
        "version",
        "roles",
        "_group_relationships",
        "__groups",
    )

    def __init__(self, http: HTTPClient, payload: UserResponse) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        relationships: list[RelationshipResponse] = self._data.pop("relationships", [])
        self.id: str = self._data["id"]
        self.username: str = self._attributes["username"]
        self.version: int = self._attributes["version"]
        self.roles: list[str] = self._attributes["roles"]
        self._group_relationships: list[ScanlationGroupResponse] = relationship_finder(
            relationships, "scanlation_group", limit=None
        )
        self.__groups: Optional[list[ScanlatorGroup]] = None

    def __repr__(self) -> str:
        return f"<User id={self.id!r} username={self.username!r}>"

    def __str__(self) -> str:
        return self.username

    def __eq__(self, other: User) -> bool:
        return isinstance(other, User) and self.id == other.id

    def __ne__(self, other: User) -> bool:
        return not self.__eq__(other)

    @property
    def url(self) -> str:
        """The URL to this user.

        Returns
        --------
        :class:`str`
            The URL of the user.
        """
        return f"https://mangadex.org/user/{self.id}"

    async def get_scanlator_groups(self) -> Optional[list[ScanlatorGroup]]:
        """|coro|

        This method will fetch the scanlator groups of this user from the API.

        Returns
        --------
        Optional[List[:class:`~hondana.ScanlatorGroup`]]
            The list of groups for this user, if any.
        """
        if not self._group_relationships:
            return

        ids = [r["id"] for r in self._group_relationships]

        data = await self._http._scanlation_group_list(
            limit=100, offset=0, ids=ids, name=None, focused_language=None, includes=ScanlatorGroupIncludes(), order=None
        )

        fmt: list[ScanlatorGroup] = [ScanlatorGroup(self._http, payload) for payload in data["data"]]
        if not fmt:
            return

        self.__groups = fmt
        return self.__groups

    @require_authentication
    async def delete(self) -> None:
        """|coro|

        This method will delete a user from the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            The response returned an error due to authentication failure.
        :exc:`NotFound`
            The user specified cannot be found.
        """

        await self._http._delete_user(self.id)



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import datetime
import json
import logging
import pathlib
import re
import warnings
from enum import Enum
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)
from urllib.parse import quote as _uriquote

import aiohttp
import multidict
from yarl import URL


try:
    import orjson
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True

from .errors import AuthenticationRequired


if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison
    from typing_extensions import Concatenate, ParamSpec

    from .types.artist import ArtistResponse
    from .types.author import AuthorResponse
    from .types.chapter import ChapterResponse
    from .types.cover import CoverResponse
    from .types.manga import MangaResponse
    from .types.relationship import RelationshipResponse
    from .types.scanlator_group import ScanlationGroupResponse
    from .types.user import UserResponse

C = TypeVar("C", bound="Any")
T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
if TYPE_CHECKING:
    B = ParamSpec("B")


LOGGER = logging.getLogger(__name__)

__all__ = (
    "MANGADEX_URL_REGEX",
    "MANGADEX_TIME_REGEX",
    "MISSING",
    "Route",
    "CustomRoute",
    "cached_slot_property",
    "to_json",
    "json_or_text",
    "php_query_builder",
    "deprecated",
    "to_snake_case",
    "to_camel_case",
    "get_image_mime_type",
    "as_chunks",
    "delta_to_iso",
    "iso_to_delta",
    "relationship_finder",
    "clean_isoformat",
    "MANGA_TAGS",
)

_PROJECT_DIR = pathlib.Path(__file__)
MAX_DEPTH: int = 10_000
MANGADEX_URL_REGEX = re.compile(
    r"(?:http[s]?:\/\/)?mangadex\.org\/(?P<type>title|chapter|author|tag)\/(?P<ID>[a-z0-9]{8}\-[a-z0-9]{4}\-[a-z0-9]{4}\-[a-z0-9]{4}\-[a-z0-9]{12})\/?(?P<title>.*)"
)
r"""
``r"(?:http[s]?:\/\/)?mangadex\.org\/(?P<type>title|chapter|author|tag)\/(?P<ID>[a-z0-9]{8}\-[a-z0-9]{4}\-[a-z0-9]{4}\-[a-z0-9]{4}\-[a-z0-9]{12})\/?(?P<title>.*)"``

This `regex pattern <https://docs.python.org/3/library/re.html#re-objects>`_ can be used to isolate common elements from a MangaDex URL.
This means that Manga, Chapter, Author or Tag urls can be parsed for their ``type``, ``ID`` and ``title``.
"""

MANGADEX_TIME_REGEX = re.compile(
    r"^(P(?P<days>[1-9]|[1-9][0-9])D)?(P?(?P<weeks>[1-9])W)?(P?T((?P<hours>[1-9]|1[0-9]|2[0-4])H)?((?P<minutes>[1-9]|[1-5][0-9]|60)M)?((?P<seconds>[1-9]|[1-5][0-9]|60)S)?)?$"
)
"""
``r"^(P([1-9]|[1-9][0-9])D)?(P?([1-9])W)?(P?T(([1-9]|1[0-9]|2[0-4])H)?(([1-9]|[1-5][0-9]|60)M)?(([1-9]|[1-5][0-9]|60)S)?)?$"``

This `regex pattern <https://docs.python.org/3/library/re.html#re-objects>`_ follows the ISO-8601 standard (MangaDex uses `PHP DateInterval <https://www.php.net/manual/en/dateinterval.construct.php>`_).
The pattern *is* usable but more meant as a guideline for your formatting.

It matches some things like: ``P1D2W`` (1 day, two weeks), ``P1D2WT3H4M`` (1 day, 2 weeks, 3 hours and 4 minutes)
"""


class Route:
    """A helper class for instantiating a HTTP method to MangaDex.

    Parameters
    -----------
    verb: :class:`str`
        The HTTP verb you wish to perform, e.g. ``"POST"``
    path: :class:`str`
        The prepended path to the API endpoint you with to target.
        e.g. ``"/manga/{manga_id}"``
    parameters: Any
        This is a special cased kwargs. Anything passed to these will substitute it's key to value in the `path`.
        E.g. if your `path` is ``"/manga/{manga_id}"``, and your parameters are ``manga_id="..."``, then it will expand into the path
        making ``"manga/..."``
    """

    BASE: ClassVar[str] = "https://api.mangadex.org"

    def __init__(self, verb: str, path: str, **parameters: Any) -> None:
        self.verb: str = verb
        self.path: str = path
        url = self.BASE + self.path
        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
        self.url: URL = URL(url, encoded=True)


class CustomRoute(Route):
    """A helper class for instantiating a HTTP method to download from MangaDex.

    Parameters
    -----------
    verb: :class:`str`
        The HTTP verb you wish to perform. E.g. ``"POST"``
    base: :class:`str`
        The base URL for the download path.
    path: :class:`str`
        The prepended path to the API endpoint you with to target.
        e.g. ``"/manga/{manga_id}"``
    parameters: Any
        This is a special cased kwargs. Anything passed to these will substitute it's key to value in the `path`.
        E.g. if your `path` is ``"/manga/{manga_id}"``, and your parameters are ``manga_id="..."``, then it will expand into the path
        making ``"manga/..."``
    """

    def __init__(self, verb: str, base: str, path: str, **parameters: Any) -> None:
        self.verb: str = verb
        self.base: str = base
        self.path: str = path
        url = self.base + self.path
        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
        self.url: URL = URL(url, encoded=True)


class MissingSentinel:
    def __eq__(self, _: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "..."


MISSING: Any = MissingSentinel()


class _StrEnum(Enum):  # type: ignore # we import this as needed.
    value: str

    def __str__(self) -> str:
        return self.value


## This class and subsequent decorator have been copied from Rapptz' Discord.py
## (https://github.com/Rapptz/discord.py)
## Credit goes to Rapptz and contributors


class CachedSlotProperty(Generic[T, T_co]):
    def __init__(self, name: str, function: Callable[[T], T_co]) -> None:
        self.name: str = name
        self.function: Callable[[T], T_co] = function
        self.__doc__ = getattr(function, "__doc__")

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> CachedSlotProperty[T, T_co]:
        ...

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> T_co:
        ...

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Any:
        if instance is None:
            return self

        try:
            return getattr(instance, self.name)
        except AttributeError:
            value = self.function(instance)
            setattr(instance, self.name, value)
            return value


def cached_slot_property(name: str, /) -> Callable[[Callable[[T], T_co]], CachedSlotProperty[T, T_co]]:
    def decorator(func: Callable[[T], T_co]) -> CachedSlotProperty[T, T_co]:
        return CachedSlotProperty(name, func)

    return decorator


def require_authentication(func: Callable[Concatenate[C, B], T]) -> Callable[Concatenate[C, B], T]:
    """A decorator to assure the `self` parameter of decorated methods has authentication set."""

    @wraps(func)
    def wrapper(item: C, *args: B.args, **kwargs: B.kwargs) -> T:
        if not item._http._authenticated:
            raise AuthenticationRequired("This method requires you to be authenticated to the API.")

        return func(item, *args, **kwargs)

    return wrapper


def deprecated(alternate: Optional[str] = None, /) -> Callable[[Callable[B, T]], Callable[B, T]]:
    """A decorator to mark a method as deprecated.

    Parameters
    -----------
    alternate: Optional[:class:`str`]
        The alternate method to use.
    """

    def decorator(func: Callable[B, T]) -> Callable[B, T]:
        @wraps(func)
        def wrapper(*args: B.args, **kwargs: B.kwargs) -> T:
            warnings.simplefilter("always", DeprecationWarning)  # turn off filter.
            if alternate is not None:
                fmt = "{0.__qualname__} is deprecated, use {1} instead."
            else:
                fmt = "{0.__qualname__} is deprecated."

            warnings.warn(fmt.format(func, alternate), stacklevel=2, category=DeprecationWarning)

            warnings.simplefilter("default", DeprecationWarning)  # reset filter.
            return func(*args, **kwargs)

        return wrapper

    return decorator


def calculate_limits(limit: int, offset: int, /, *, max_limit: int = 100) -> tuple[int, int]:
    """A helper function that will calculate the offset and limit parameters for API endpoints.

    Parameters
    -----------
    limit: :class:`int`
        The limit (or amount) of objects you are requesting.
    offset: :class:`int`
        The offset (or pagination start point) for the objects you are requesting.
    max_limit: :class:`int`
        The maximum limit value for the API Endpoint.

    Raises
    -------
    :exc:`ValueError`
        Exceeding the maximum pagination limit.

    Returns
    --------
    Tuple[:class:`int`, :class:`int`]
    """
    if offset >= MAX_DEPTH:
        raise ValueError(f"An offset of {MAX_DEPTH} will not return results.")

    offset = max(offset, 0)

    difference = MAX_DEPTH - offset
    if difference <= max_limit:
        new_limit = difference
        new_offset = MAX_DEPTH - new_limit
        return new_limit, new_offset

    new_limit = min(max(0, limit), max_limit)
    new_offset = min(max(0, offset), MAX_DEPTH - new_limit)

    return new_limit, new_offset


if HAS_ORJSON is True:

    def to_json(obj: Any, /) -> str:
        """A quick method that dumps a Python type to JSON object."""
        return orjson.dumps(obj).decode("utf-8")

    _from_json = orjson.loads  # type: ignore # this is guarded in an if.

else:

    def to_json(obj: Any, /) -> str:
        """A quick method that dumps a Python type to JSON object."""
        return json.dumps(obj, separators=(",", ":"), ensure_ascii=True)

    _from_json = json.loads


async def json_or_text(response: aiohttp.ClientResponse, /) -> Union[dict[str, Any], str]:
    """A quick method to parse a `aiohttp.ClientResponse` and test if it's json or text."""
    text = await response.text(encoding="utf-8")
    try:
        if response.headers["content-type"] == "application/json":
            try:
                return _from_json(text)
            except json.JSONDecodeError:
                pass
    except KeyError:
        pass

    return text


def php_query_builder(
    obj: Mapping[str, Optional[Union[str, int, bool, list[str], dict[str, str]]]], /
) -> multidict.MultiDict[Union[str, int]]:
    """
    A helper function that builds a MangaDex (PHP) query string from a mapping.

    Parameters
    -----------
    obj: Mapping[:class:`str`, Optional[Union[:class:`str`, :class:`int`, :class:`bool`, List[:class:`str`], Dict[:class:`str`, :class:`str`]]]]
        The mapping to build the query string from.
    """
    fmt = multidict.MultiDict[Union[str, int]]()
    for key, value in obj.items():
        if value is None:
            fmt.add(key, "null")
        elif isinstance(value, str):
            fmt.add(key, value)
        elif isinstance(value, bool):
            fmt.add(key, str(value).lower())
        elif isinstance(value, list):
            for item in value:
                fmt.add(f"{key}[]", item)
        elif isinstance(value, dict):
            for subkey, subvalue in value.items():
                fmt.add(f"{key}[{subkey}]", subvalue)
        else:
            fmt.add(key, value)

    return fmt


def get_image_mime_type(data: bytes, /) -> str:
    """Returns the image type from the first few bytes."""
    if data.startswith(b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif data[:3] == b"\xff\xd8\xff" or data[6:10] in (b"JFIF", b"Exif"):
        return "image/jpeg"
    elif data.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"
    elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    else:
        raise ValueError("Unsupported image type given")


def to_snake_case(string: str, /) -> str:
    """Quick function to return snake_case from camelCase."""
    fmt: list[str] = []
    for character in string:
        if character.isupper():
            fmt.append(f"_{character.lower()}")
            continue
        fmt.append(character)
    return "".join(fmt)


def to_camel_case(string: str, /) -> str:
    """Quick function to return camelCase from snake_case."""
    first, *rest = string.split("_")
    chunks = [first.lower(), *map(str.capitalize, rest)]

    return "".join(chunks)


def as_chunks(iterator: Iterable[T], /, max_size: int) -> Iterable[list[T]]:
    ret = []
    n = 0
    for item in iterator:
        ret.append(item)
        n += 1
        if n == max_size:
            yield ret
            ret = []
            n = 0
    if ret:
        yield ret


def delta_to_iso(delta: datetime.timedelta, /) -> str:
    """A helper method to dump a timedelta to an ISO 8601 timedelta string.

    Parameters
    -----------
    delta: :class:`datetime.timedelta`
        The timedelta to convert.

    Returns
    --------
    :class:`str`
        The converted string.
    """
    seconds = round(delta.total_seconds())
    weeks, seconds = divmod(seconds, 60 * 60 * 24 * 7)
    days, seconds = divmod(seconds, 60 * 60 * 24)
    hours, seconds = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(seconds, 60)

    builder = "P" * bool(weeks or days)
    if days:
        builder += f"{days}D"
    if weeks:
        builder += f"{weeks}W"

    builder += "T" * bool(hours or minutes or seconds)

    if hours:
        builder += f"{hours}H"
    if minutes:
        builder += f"{minutes}M"
    if seconds:
        builder += f"{seconds}S"

    return builder


def iso_to_delta(iso: str, /) -> datetime.timedelta:
    """A helper method to load a timedelta from an ISO8601 string.

    Parameters
    -----------
    iso: :class:`str`
        The ISO8601 datetime string to parse.

    Raises
    -------
    :exc:`TypeError`
        If the given string is not a valid ISO8601 string and does not match :class:`~hondana.utils.MANGADEX_TIME_REGEX`.

    Returns
    --------
    :class:`datetime.timedelta`
        The timedelta based on the parsed string.
    """
    if (match := MANGADEX_TIME_REGEX.fullmatch(iso)) is None:
        raise TypeError("The passed string does not match the regex pattern.")

    match_dict = match.groupdict()

    times: dict[str, float] = {key: float(value) for key, value in match_dict.items() if value}
    return datetime.timedelta(**times)


RelType = Literal[
    "artist",
    "author",
    "chapter",
    "cover_art",
    "leader",
    "manga",
    "member",
    "scanlation_group",
    "user",
]


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["artist"], *, limit: Literal[1]
) -> ArtistResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["artist"], *, limit: Optional[int] = ...
) -> list[ArtistResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["author"], *, limit: Literal[1]
) -> AuthorResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["author"], *, limit: Optional[int] = ...
) -> list[AuthorResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["chapter"], *, limit: Literal[1]
) -> ChapterResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["chapter"], *, limit: Optional[int] = ...
) -> list[ChapterResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["cover_art"], *, limit: Literal[1]
) -> CoverResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["cover_art"], *, limit: Optional[int] = ...
) -> list[CoverResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["leader"], *, limit: Literal[1]
) -> UserResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["leader"], *, limit: Optional[int] = ...
) -> list[UserResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["manga"], *, limit: Literal[1]
) -> MangaResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["manga"], *, limit: Optional[int] = ...
) -> list[MangaResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["member"], *, limit: Literal[1]
) -> UserResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["member"], *, limit: Optional[int] = ...
) -> list[UserResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["scanlation_group"], *, limit: Literal[1]
) -> ScanlationGroupResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["scanlation_group"], *, limit: Optional[int] = ...
) -> list[ScanlationGroupResponse]:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["user"], *, limit: Literal[1]
) -> UserResponse:
    ...


@overload
def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: Literal["user"], *, limit: Optional[int] = ...
) -> list[UserResponse]:
    ...


def relationship_finder(
    relationships: list[RelationshipResponse], relationship_type: RelType, *, limit: Optional[int] = None
) -> Optional[Union[list[Any], Any]]:
    if not relationships:
        return

    ret: list[RelationshipResponse] = []
    relationships_copy = relationships.copy()

    for relationship in relationships_copy:
        if relationship["type"] == relationship_type:
            if limit == 1:
                relationships.remove(relationship)
                return relationship
            else:
                ret.append(relationship)
            relationships.remove(relationship)

    if limit is not None:
        return ret[:limit]

    return ret


def clean_isoformat(dt: datetime.datetime, /) -> str:
    """A helper method to cleanly convert a datetime (aware or naive) to a timezone-less ISO8601 string.

    Parameters
    -----------
    dt: :class:`datetime.datetime`
        The datetime to convert.

    Returns
    --------
    :class:`str`
        The ISO8601 string.


    .. note::
        The passed datetime will have its timezone converted to UTC and then have its timezone stripped.

    """
    if dt.tzinfo != datetime.timezone.utc:
        dt = dt.astimezone(datetime.timezone.utc)

    if dt.tzinfo is not None:
        dt = dt.replace(tzinfo=None)

    return dt.isoformat(timespec="seconds")


_PATH_WITH_EXTRA = re.compile(r"(?P<num>\d+)(\-?(?P<extra>\w*))?\.(?P<ext>png|jpg|gif|webm)")


def upload_file_sort(key: SupportsRichComparison) -> tuple[int, str]:
    if isinstance(key, pathlib.Path):
        if match := _PATH_WITH_EXTRA.fullmatch(key.name):
            return (len(match["num"]), match["num"])

    raise ValueError("Invalid filename format given.")


_tags_path: pathlib.Path = _PROJECT_DIR.parent / "extras" / "tags.json"
with _tags_path.open("r") as _tags_fp:
    MANGA_TAGS: dict[str, str] = json.load(_tags_fp)


class _ReportReasons(TypedDict):
    manga: dict[str, str]
    chapter: dict[str, str]
    scanlation_group: dict[str, str]
    author: dict[str, str]
    user: dict[str, str]


_report_reason_path = _PROJECT_DIR.parent / "extras" / "report_reasons.json"
with _report_reason_path.open("r") as _reports_fp:
    _REPORT_REASONS: _ReportReasons = json.load(_reports_fp)



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal, Optional

from .relationship import Relationship
from .utils import MANGA_TAGS, cached_slot_property


if TYPE_CHECKING:
    from .types.common import LocalizedString
    from .types.relationship import RelationshipResponse
    from .types.tags import TagResponse


__all__ = (
    "Tag",
    "QueryTags",
)

logger: logging.Logger = logging.getLogger("hondana")


class Tag:
    """A class representing a single Tag from MangaDex.

    Attributes
    -----------
    id: :class:`str`
        The UUID associated with this tag.
    group: :class:`str`
        The group (or kind) of tag.
        e.g. if it is a genre specification, or theme.
    version: :class:`int`
        Version revision of this tag.


    .. note::
        The tag descriptions are currently empty, but seemingly they will be localised descriptions of each one.
        i.e. ``[{"en": "some tag description"}]``

    .. note::
        All tag names currently only have the ``"en"`` key attributed to their localization, so we return this by default.
    """

    __slots__ = (
        "_data",
        "_attributes",
        "_relationships",
        "_name",
        "_description",
        "id",
        "group",
        "version",
        "_cs_relationships",
    )

    def __init__(self, payload: TagResponse) -> None:
        self._data = payload
        self._attributes = payload["attributes"]
        self._relationships: list[RelationshipResponse] = self._data.pop(  # type: ignore # can't pop from a TypedDict
            "relationships", []  # type: ignore # can't pop from a TypedDict
        )  # TODO: remove this when they have relationships to be in line.
        self._name = self._attributes["name"]
        self.id: str = payload["id"]
        self._description: LocalizedString = self._attributes["description"]
        self.group: str = self._attributes["group"]
        self.version: int = self._attributes["version"]

    def __repr__(self) -> str:
        return f"<Tag id={self.id!r} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: Tag) -> bool:
        return isinstance(other, Tag) and self.id == other.id

    def __ne__(self, other: Tag) -> bool:
        return not self.__eq__(other)

    @property
    def name(self) -> str:
        """The name of the tag.

        Returns
        --------
        :class:`str`
        """
        name = self._name.get("en")
        if name is None:
            key = next(iter(self._name))
            return self._name[key]
        return name

    @property
    def description(self) -> str | None:
        """The description of the tag, if any.

        Returns
        --------
        Optional[:class:`str`]
        """
        if not self._description:
            return None

        description = self._description.get("en")
        if description is None:
            key = next(iter(self._description))
            return self._description[key]
        return description

    @property
    def url(self) -> str:
        """The URL to this tag.
        Allows browsing all manga attributed to this tag

        Returns
        --------
        :class:`str`
            The URL of the tag.
        """
        return f"https://mangadex.org/tag/{self.id}"

    @cached_slot_property("_cs_relationships")
    def relationships(self) -> list[Relationship]:
        """The relationships of this Tag.

        Returns
        --------
        List[:class:`~hondana.Relationship`]
            The list of relationships this tag has.
        """
        return [Relationship(item) for item in self._relationships]


class QueryTags:
    """Utility class for creating a Tag based query.

    Attributes
    -----------
    tags: List[:class:`str`]
        The tag names you are going to query.
    mode: Literal[``"AND"``, ``"OR"``]
        The logical operation for the tags.

    Raises
    -------
    :exc:`TypeError`
        The mode specified for the builder is not one of "AND" or "OR"
    :exc:`ValueError`
        The tags passed did not align with MangaDex tags.


    .. note::
        The tags passed need to match the *local* cache of the tags.
        If you feel this is out of date, you can try the helper method :meth:`~hondana.Client.update_tags`
    """

    def __init__(self, *tags: str, mode: Literal["AND", "OR"] = "AND"):
        self._tags = tags
        self.tags: Optional[list[str]] = None
        self.mode: str = mode.upper()
        if self.mode not in {"AND", "OR"}:
            raise TypeError("Tags mode has to be 'AND' or 'OR'.")
        self._set_tags()

    def __repr__(self) -> str:
        return f"<Tags mode={self.mode} number_of_tags={len(self._tags)}>"

    def _set_tags(self) -> list[str]:
        tags = []
        for tag in self._tags:
            if tag_ := MANGA_TAGS.get(tag.title()):
                tags.append(tag_)
            else:
                logger.warning("Tag '%s' cannot be found in the local tag cache, skipping.", tag)

        if not tags:
            raise ValueError("No tags passed matched any valid MangaDex tags.")

        self.tags = tags
        return self.tags



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Optional, Union

from .artist import Artist
from .author import Author
from .collections import ChapterFeed, MangaRelationCollection
from .cover import Cover
from .enums import ContentRating, MangaRelationType, MangaState, MangaStatus, PublicationDemographic, ReadingStatus
from .query import ArtistIncludes, AuthorIncludes, ChapterIncludes, CoverIncludes, FeedOrderQuery, MangaIncludes
from .tags import Tag
from .types.cover import CoverResponse
from .utils import MISSING, cached_slot_property, relationship_finder, require_authentication


if TYPE_CHECKING:
    from .http import HTTPClient
    from .tags import QueryTags
    from .types import manga
    from .types.artist import ArtistResponse
    from .types.author import AuthorResponse
    from .types.common import LanguageCode, LocalizedString
    from .types.manga import MangaResponse
    from .types.relationship import RelationshipResponse
    from .types.statistics import BatchStatisticsResponse, PersonalMangaRatingsResponse, StatisticsResponse


__all__ = (
    "Manga",
    "MangaRelation",
    "MangaStatistics",
    "MangaRating",
)


class Manga:
    """A class representing a Manga returned from the MangaDex API.

    Attributes
    -----------
    id: :class:`str`
        The UUID associated to this manga.
    relation_type: Optional[:class:`~hondana.MangaRelationType`]
        The type of relation this is, to the parent manga requested.
        Only available when :meth:`get_related_manga` is called.
    alternate_titles: :class:`~hondana.types.common.LocalizedString`
        The alternative title mapping for the Manga.
        i.e. ``{"en": "Some Other Title"}``
    locked: :class:`bool`
        If the Manga is considered 'locked' or not in the API.
    links: :class:`~hondana.types.manga.MangaLinks`
        The mapping of links the API has attributed to the Manga.
        (see: https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data)
    original_language: :class:`str`
        The language code for the original language of the Manga.
    last_volume: Optional[:class:`str`]
        The last volume attributed to the manga, if any.
    last_chapter: Optional[:class:`str`]
        The last chapter attributed to the manga, if any.
    publication_demographic: Optional[:class:`~hondana.PublicationDemographic`]
        The attributed publication demographic(s) for the manga, if any.
    year: Optional[:class:`int`]
        The year the manga was release, if the key exists.
    content_rating: Optional[:class:`~hondana.ContentRating`]
        The content rating attributed to the manga, if any.
    chapter_numbers_reset_on_new_volume: :class:`bool`
        Whether the chapter numbers will reset on a new volume or not.
    latest_uploaded_chapter: :class:`str`
        The ID of the latest uploaded chapter of this manga.
    state: Optional[:class:`~hondana.MangaState`]
        The publication state of the Manga.
    stats: Optional[:class:`~hondana.MangaStatistics`]
        The statistics of the manga.
    version: :class:`int`
        The version revision of this manga.


    .. note::
        The :attr:`stats` is only populated after :meth:`~hondana.Manga.get_statistics` is called.
    """

    __slots__ = (
        "_http",
        "_data",
        "_attributes",
        "_title",
        "_description",
        "id",
        "relation_type",
        "alternate_titles",
        "locked",
        "links",
        "original_language",
        "last_volume",
        "last_chapter",
        "publication_demographic",
        "status",
        "year",
        "content_rating",
        "chapter_numbers_reset_on_new_volume",
        "available_translated_languages",
        "latest_uploaded_chapter",
        "state",
        "stats",
        "version",
        "_tags",
        "_created_at",
        "_updated_at",
        "_artist_relationships",
        "_author_relationships",
        "_related_manga_relationships",
        "_cover_relationship",
        "__authors",
        "__artists",
        "__cover",
        "__related_manga",
        "_cs_tags",
    )

    def __init__(self, http: HTTPClient, payload: manga.MangaResponse) -> None:
        self._http = http
        self._data = payload
        relationships: list[RelationshipResponse] = self._data.pop("relationships", [])
        self._attributes = payload["attributes"]
        self.id: str = payload["id"]
        self._title = self._attributes["title"]
        self._description: LocalizedString = self._attributes["description"]
        _related = payload.get("related", None)
        self.relation_type: Optional[MangaRelationType] = MangaRelationType(_related) if _related else None
        self.alternate_titles: LocalizedString = {k: v for item in self._attributes["altTitles"] for k, v in item.items()}  # type: ignore # this is actually valid but breaks pylance (not pyright...?)? TODO: test in later version
        self.locked: bool = self._attributes.get("isLocked", False)
        self.links: manga.MangaLinks = self._attributes["links"]
        self.original_language: str = self._attributes["originalLanguage"]
        self.last_volume: Optional[str] = self._attributes["lastVolume"]
        self.last_chapter: Optional[str] = self._attributes["lastChapter"]
        self.publication_demographic: Optional[PublicationDemographic] = (
            PublicationDemographic(self._attributes["publicationDemographic"])
            if self._attributes["publicationDemographic"]
            else None
        )
        self.status: Optional[MangaStatus] = MangaStatus(self._attributes["status"])
        self.year: Optional[int] = self._attributes["year"]
        self.content_rating: Optional[ContentRating] = (
            ContentRating(self._attributes["contentRating"]) if self._attributes["contentRating"] else None
        )
        self.chapter_numbers_reset_on_new_volume: bool = self._attributes["chapterNumbersResetOnNewVolume"]
        self.available_translated_languages: list[LanguageCode] = self._attributes["availableTranslatedLanguages"]
        self.latest_uploaded_chapter: str = self._attributes["latestUploadedChapter"]
        self.state: Optional[MangaState] = MangaState(self._attributes["state"]) if self._attributes["state"] else None
        self.stats: Optional[MangaStatistics] = None
        self.version: int = self._attributes["version"]
        self._tags = self._attributes["tags"]
        self._created_at = self._attributes["createdAt"]
        self._updated_at = self._attributes["updatedAt"]
        self._author_relationships: list[AuthorResponse] = relationship_finder(relationships, "author", limit=None)
        self._artist_relationships: list[ArtistResponse] = relationship_finder(relationships, "artist", limit=None)
        self._related_manga_relationships: list[MangaResponse] = relationship_finder(relationships, "manga", limit=None)
        self._cover_relationship: Optional[CoverResponse] = relationship_finder(relationships, "cover_art", limit=1)
        self.__authors: Optional[list[Author]] = None
        self.__artists: Optional[list[Artist]] = None
        self.__cover: Optional[Cover] = None
        self.__related_manga: Optional[list[Manga]] = None

    def __repr__(self) -> str:
        return f"<Manga id={self.id!r} title={self.title!r}>"

    def __str__(self) -> str:
        return self.title

    def __eq__(self, other: Union[Manga, MangaRelation]) -> bool:
        return self.id == other.id

    def __ne__(self, other: Union[Manga, MangaRelation]) -> bool:
        return not self.__eq__(other)

    @property
    def url(self) -> str:
        """The URL to this manga.

        Returns
        --------
        :class:`str`
            The URL of the manga.
        """
        return f"https://mangadex.org/title/{self.id}"

    @property
    def title(self) -> str:
        """The manga's title.

        Returns
        --------
        :class:`str`
            The title of the manga, defaults to the ``en`` key in the titles attribute of the response.
            Attempts to fall back to the manga's default language, and failing that it will
            fall back to the next available key if ``en`` or default language key is not present.
        """
        title = self._title.get("en")
        if title is None:
            key = next(iter(self._title))
            title = self._title.get(self.original_language, self._title[key])

        return title

    @property
    def description(self) -> Optional[str]:
        """The manga's description/synopsis.

        Returns
        --------
        Optional[:class:`str`]
            The description of the manga, defaults to the ``en`` key in the titles.
            Falls back to the next available key if ``en`` is not present. If there is
            no description, returns None.
        """
        desc = self._description.get("en")
        if desc is None:
            if not self._description:
                return None
            key = next(iter(self._description))
            return self._description[key]
        return desc

    @property
    def raw_description(self) -> LocalizedString:
        """The raw description attribute from the manga's payload from the API.

        Returns
        --------
        :class:`~hondana.types.common.LocalizedString`
            The raw object from the manga's api response payload.
            Provides no formatting on its own. Consider :meth:`~hondana.Manga.description` or :meth:`~hondana.Manga.localised_description` instead.
        """
        return self._description

    @property
    def alt_titles(self) -> LocalizedString:
        """The raw alt_titles attribute from the manga's payload from the API.

        Returns
        --------
        :class:`hondana.types.common.LocalizedString`
            The raw object from the manga's payload.
            Provides no formatting on its own. Consider :meth:`~hondana.Manga.localised_title` instead.
        """
        return self.alternate_titles

    @property
    def created_at(self) -> datetime.datetime:
        """The date this manga was created.

        Returns
        --------
        :class:`datetime.datetime`
            The UTC datetime of when this manga was created.
        """
        return datetime.datetime.fromisoformat(self._created_at)

    @property
    def updated_at(self) -> datetime.datetime:
        """The date this manga was last updated.

        Returns
        --------
        :class:`datetime.datetime`
            The UTC datetime of when this manga was last updated.
        """
        return datetime.datetime.fromisoformat(self._updated_at)

    @cached_slot_property("_cs_tags")
    def tags(self) -> list[Tag]:
        """The tags of this Manga.

        Returns
        --------
        List[:class:`~hondana.Tag`]
            The list of tags that this manga is associated with.
        """
        return [Tag(item) for item in self._tags]

    @property
    def artists(self) -> Optional[list[Artist]]:
        """The artists of the parent Manga.


        .. note::
            If the parent manga was **not** requested with the "artist" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        --------
        Optional[List[:class:`~hondana.Artist`]]
            The artists associated with this Manga.
        """
        if self.__artists is not None:
            return self.__artists

        if not self._artist_relationships:
            return

        formatted: list[Artist] = [
            Artist(self._http, artist) for artist in self._artist_relationships if "attributes" in artist
        ]

        if not formatted:
            return None

        self.__artists = formatted
        return self.__artists

    @artists.setter
    def artists(self, value: list[Artist]) -> None:
        self.__artists = [item for item in value if isinstance(item, Artist)]

    @property
    def authors(self) -> Optional[list[Author]]:
        """The artists of the parent Manga.


        .. note::
            If the parent manga was **not** requested with the "artist" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        --------
        Optional[List[:class:`~hondana.Author`]]
            The artists associated with this Manga.

        """
        if self.__authors is not None:
            return self.__authors

        if not self._author_relationships:
            return

        formatted: list[Author] = [
            Author(self._http, author) for author in self._author_relationships if "attributes" in author
        ]

        if not formatted:
            return

        self.__authors = formatted
        return self.__authors

    @authors.setter
    def authors(self, value: list[Author]) -> None:
        self.__authors = [item for item in value if isinstance(item, Author)]

    @property
    def cover(self) -> Optional[Cover]:
        """The cover of the manga.


        .. note::
            If the parent manga was **not** requested with the "cover" `includes[]` query parameter
            then this method will return ``None``.

        Returns
        --------
        Optional[:class:`~hondana.Cover`]
            The cover of the manga.
        """
        if self.__cover is not None:
            return self.__cover

        if not self._cover_relationship:
            return

        if "attributes" in self._cover_relationship:
            self.__cover = Cover(self._http, self._cover_relationship)
            return self.__cover

    @cover.setter
    def cover(self, value: Cover) -> None:
        if isinstance(value, Cover):
            self.__cover = value

    @property
    def related_manga(self) -> Optional[list[Manga]]:
        """The related manga of the parent Manga.

        Returns
        --------
        Optional[List[:class:`~hondana.Manga`]]
            The related manga of the parent manga.
        """
        if self.__related_manga is not None:
            return self.__related_manga

        if not self._related_manga_relationships:
            return

        formatted: list[Manga] = [
            self.__class__(self._http, item) for item in self._related_manga_relationships if "attributes" in item
        ]

        if not formatted:
            return

        self.__related_manga = formatted
        return self.__related_manga

    @related_manga.setter
    def related_manga(self, value: list[Manga]) -> None:
        self.__related_manga = [item for item in value if isinstance(item, self.__class__)]

    async def get_artists(self) -> Optional[list[Artist]]:
        """|coro|

        This method will return the artists of the manga and caches the response.


        .. note::
            If the parent manga was requested with the "artist" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve the artist data.

        Returns
        --------
        Optional[List[:class:`~hondana.Author`]]
            The artists of the manga.
        """
        if self.artists is not None:
            return self.artists

        if not self._artist_relationships:
            return

        ids = [r["id"] for r in self._artist_relationships]

        formatted: list[Artist] = []
        for item in ids:
            data = await self._http._get_artist(item, includes=ArtistIncludes())
            formatted.append(Artist(self._http, data["data"]))

        if not formatted:
            return

        self.artists = formatted
        return formatted

    async def get_authors(self) -> Optional[list[Author]]:
        """|coro|

        This method will return the authors of the manga and caches the response.


        .. note::
            If the parent manga was requested with the "author" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve the author data.

        Returns
        --------
        Optional[List[:class:`~hondana.Author`]]
            The authors of the manga.
        """
        if self.authors is not None:
            return self.authors

        if not self._author_relationships:
            return

        ids = [r["id"] for r in self._related_manga_relationships]

        formatted: list[Author] = []
        for item in ids:
            data = await self._http._get_author(item, includes=AuthorIncludes())
            formatted.append(Author(self._http, data["data"]))

        if not formatted:
            return

        self.authors = formatted
        return formatted

    async def get_cover(self) -> Optional[Cover]:
        """|coro|

        This method will return the cover URL of the parent Manga if it exists and caches the response.

        Returns
        --------
        Optional[:class:`~hondana.Cover`]
            The Cover associated with this Manga.
        """
        if self.cover is not None:
            return self.cover

        if not self._cover_relationship:
            return

        data = await self._http._get_cover(self._cover_relationship["id"], includes=CoverIncludes())
        self.cover = Cover(self._http, data["data"])
        return self.cover

    def cover_url(self, *, size: Optional[Literal[256, 512]] = None) -> Optional[str]:
        """This method will return a direct url to the cover art of the parent Manga.

        If the manga was requested without the ``"cover_art"`` includes[] parameters, then this method will return ``None``.


        .. note::
            For a more stable cover return, try :meth:`~hondana.Manga.get_cover`
        """
        if not self.cover:
            return

        return self.cover.url(size, parent_id=self.id)

    async def get_related_manga(self) -> Optional[list[Manga]]:
        """|coro|

        This method will return all the related manga and cache their response.


        .. note::
            If the parent manga was requested with the "manga" `includes[]` query parameter,
            then this method will not make extra API calls to retrieve manga data.

        Returns
        --------
        Optional[List[:class:`~hondana.Manga`]]
            The related manga of the parent.
        """
        if self.related_manga is not None:
            return self.related_manga

        if not self._related_manga_relationships:
            return

        ids = [r["id"] for r in self._related_manga_relationships]

        data = await self._http._manga_list(
            limit=100,
            offset=0,
            title=None,
            authors=None,
            artists=None,
            year=None,
            included_tags=None,
            excluded_tags=None,
            status=None,
            original_language=None,
            excluded_original_language=None,
            available_translated_language=None,
            publication_demographic=None,
            ids=ids,
            content_rating=None,
            created_at_since=None,
            updated_at_since=None,
            order=None,
            includes=MangaIncludes(),
            has_available_chapters=None,
            group=None,
        )

        ret: list[Manga] = [Manga(self._http, item) for item in data["data"]]
        if not ret:
            return

        self.related_manga = ret
        return self.related_manga

    @require_authentication
    async def update(
        self,
        *,
        title: Optional[LocalizedString] = None,
        alt_titles: Optional[list[LocalizedString]] = None,
        description: Optional[LocalizedString] = None,
        authors: Optional[list[str]] = None,
        artists: Optional[list[str]] = None,
        links: Optional[manga.MangaLinks] = None,
        original_language: Optional[str] = None,
        last_volume: Optional[str] = MISSING,
        last_chapter: Optional[str] = MISSING,
        publication_demographic: PublicationDemographic = MISSING,
        status: Optional[MangaStatus] = None,
        year: Optional[int] = MISSING,
        content_rating: Optional[ContentRating] = None,
        tags: Optional[QueryTags] = None,
        primary_cover: Optional[str] = MISSING,
        version: int,
    ) -> Manga:
        """|coro|

        This method will update the current Manga within the MangaDex API.

        Parameters
        -----------
        title: Optional[:class:`~hondana.types.common.LocalizedString`]
            The manga titles in the format of ``language_key: title``
            i.e. ``{"en": "Some Manga Title"}``
        alt_titles: Optional[List[:class:`~hondana.types.common.LocalizedString`]]
            The alternative titles in the format of ``language_key: title``
            i.e. ``[{"en": "Some Other Title"}, {"fr": "Un Autre Titre"}]``
        description: Optional[:class:`~hondana.types.common.LocalizedString`]
            The manga description in the format of ``language_key: description``
            i.e. ``{"en": "My amazing manga where x y z happens"}``
        authors: Optional[List[:class:`str`]]
            The list of author UUIDs to credit to this manga.
        artists: Optional[List[:class:`str`]]
            The list of artist UUIDs to credit to this manga.
        links: Optional[:class:`~hondana.types.manga.MangaLinks`]
            The links relevant to the manga.
            See here for more details: https://api.mangadex.org/docs.html#section/Static-data/Manga-links-data
        original_language: Optional[:class:`str`]
            The language key for the original language of the manga.
        last_volume: Optional[:class:`str`]
            The last volume to attribute to this manga.
        last_chapter: Optional[:class:`str`]
            The last chapter to attribute to this manga.
        publication_demographic: Optional[:class:`~hondana.PublicationDemographic`]
            The target publication demographic of this manga.
        status: Optional[:class:`~hondana.MangaStatus`]
            The status of the manga.
        year: Optional[:class:`int`]
            The release year of the manga.
        content_rating: Optional[:class:`~hondana.ContentRating`]
            The content rating of the manga.
        tags: Optional[:class:`QueryTags`]
            The QueryTags instance for the list of tags to attribute to this manga.
        primmary_cover: Optional[:class:`str`]
            The UUID representing the cover that should show for this manga as it's primary.
        version: :class:`int`
            The revision version of this manga.


        .. note::
            The ``last_volume``, ``last_chapter``, ``publication_demographic``, ``year`` and ``primary_cover`` parameters
            are nullable in the API. Pass ``None`` explicitly to do so.

        Raises
        -------
        :exc:`BadRequest`
            The query parameters were not valid.
        :exc:`Forbidden`
            The update returned an error due to authentication failure.
        :exc:`NotFound`
            The specified manga does not exist.

        Returns
        --------
        :class:`~hondana.Manga`
            The manga that was returned after creation.
        """
        data = await self._http._update_manga(
            self.id,
            title=title,
            alt_titles=alt_titles,
            description=description,
            authors=authors,
            artists=artists,
            links=links,
            original_language=original_language,
            last_volume=last_volume,
            last_chapter=last_chapter,
            publication_demographic=publication_demographic,
            status=status,
            year=year,
            content_rating=content_rating,
            tags=tags,
            primary_cover=primary_cover,
            version=version,
        )

        return self.__class__(self._http, data["data"])

    @require_authentication
    async def delete(self) -> None:
        """|coro|

        This method will delete a Manga from the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            The update returned an error due to authentication failure.
        :exc:`NotFound`
            The specified manga does not exist.
        """

        await self._http._delete_manga(self.id)

    @require_authentication
    async def unfollow(self) -> None:
        """|coro|

        This method will unfollow the current Manga for the logged-in user in the MangaDex API.

        Raises
        -------
        :exc:`Forbidden`
            The request returned an error due to authentication failure.
        :exc:`NotFound`
            The specified manga does not exist.
        """
        await self._http._unfollow_manga(self.id)

    @require_authentication
    async def follow(self, *, set_status: bool = True, status: ReadingStatus = ReadingStatus.reading) -> None:
        """|coro|

        This method will follow the current Manga for the logged-in user in the MangaDex API.

        Parameters
        -----------
        set_status: :class:`bool`
            Whether to set the reading status of the manga you follow.
            Due to the current MangaDex infrastructure, not setting a status will cause the manga to not show up in your lists.
            Defaults to ``True``
        status: :class:`~hondana.ReadingStatus`
            The status to apply to the newly followed manga.
            Irrelevant if ``set_status`` is ``False``.

        Raises
        -------
        :exc:`Forbidden`
            The request returned an error due to authentication failure.
        :exc:`NotFound`
            The specified manga does not exist.
        """
        await self._http._follow_manga(self.id)
        if set_status:
            await self.update_reading_status(status=status)

    def localised_title(self, language_code: LanguageCode, /) -> Optional[str]:
        """
        This method will attempt to return the current manga's title in the provided language code.
        Falling back to the :attr:`title`.

        Aliased to :meth:`~Manga.localized_title`

        Parameters
        -----------
        language_code: :class:`~hondana.types.common.LanguageCode`
            The language code to attempt to return the manga name in.

        Returns
        --------
        Optional[:class:`str`]
            The manga name in the provided language, if found.
        """
        return self.alternate_titles.get(language_code, self.title)

    localized_title = localised_title

    def localised_description(self, language_code: LanguageCode, /) -> Optional[str]:
        """
        This method will attempt to return the current manga's description in the provided language code.
        Falling back to the :attr:`description`.

        Aliased to :meth:`~Manga.localized_description`

        Parameters
        -----------
        language_code: :class:`~hondana.types.common.LanguageCode`
            The language code to attempt to return the manga description in.

        Returns
        --------
        Optional[:class:`str`]
            The manga description in the provided language, if found.
        """
        return self._description.get(language_code, self.description)

    localized_description = localised_description

    async def feed(
        self,
        *,
        limit: Optional[int] = 100,
        offset: int = 0,
        translated_language: Optional[list[LanguageCode]] = None,
        original_language: Optional[list[LanguageCode]] = None,
        excluded_original_language: Optional[list[LanguageCode]] = None,
        content_rating: Optional[list[ContentRating]] = None,
        excluded_groups: Optional[list[str]] = None,
        excluded_uploaders: Optional[list[str]] = None,
        include_future_updates: Optional[bool] = None,
        created_at_since: Optional[datetime.datetime] = None,
        updated_at_since: Optional[datetime.datetime] = None,
        published_at_since: Optional[datetime.datetime] = None,
        order: Optional[FeedOrderQuery] = None,
        includes: Optional[ChapterIncludes] = ChapterIncludes(),
        include_empty_pages: Optional[bool] = None,
        include_future_publish_at: Optional[bool] = None,
        include_external_url: Optional[bool] = None,
    ) -> ChapterFeed:
        """|coro|

        This method returns the current manga's chapter feed.

        Parameters
        -----------
        limit: Optional[:class:`int`]
            Defaults to 100. The maximum amount of chapters to return-in the response.
        offset: :class:`int`
            Defaults to 0. The pagination offset for the request.
        translated_language: List[:class:`str`]
            A list of language codes to filter the returned chapters with.
        original_language: List[:class:`~hondana.types.common.LanguageCode`]
            A list of language codes to filter the original language of the returned chapters with.
        excluded_original_language: List[:class:`~hondana.types.common.LanguageCode`]
            A list of language codes to negate filter the original language of the returned chapters with.
        content_rating: Optional[List[:class:`~hondana.ContentRating`]]
            The content rating to filter the feed by.
        excluded_groups: Optional[List[:class:`str`]]
            The list of scanlator groups to exclude from the response.
        excluded_uploaders: Optional[List[:class:`str`]]
            The list of uploaders to exclude from the response.
        include_future_updates: Optional[:class:`bool`]
            Whether to include future chapters in this feed. Defaults to ``True`` API side.
        created_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their creation date.
        updated_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their updated at date.
        published_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their published at date.
        order: Optional[:class:`~hondana.query.FeedOrderQuery`]
            A query parameter to choose how the responses are ordered.
            i.e. ``{"chapters": "desc"}``
        includes: Optional[:class:`~hondana.query.ChapterIncludes`]
            The list of options to include increased payloads for per chapter.
            Defaults to these values.
        include_empty_pages: Optional[:class:`bool`]
            Whether to show chapters with no pages available.
        include_future_publish_at: Optional[:class:`bool`]
            Whether to show chapters with a publishAt value set in the future.
        includeExternalUrl: Optional[:class:`bool`]
            Whether to show chapters that have an external URL attached to them.


        .. note::
            Passing ``None`` to ``limit`` will attempt to retrieve all items in the chapter feed.

        Raises
        -------
        :exc:`BadRequest`
            The query parameters were malformed.

        Returns
        --------
        :class:`~hondana.ChapterFeed`
            Returns a collection of chapters.
        """

        inner_limit = limit or 100

        chapters = []
        while True:
            data = await self._http._manga_feed(
                self.id,
                limit=inner_limit,
                offset=offset,
                translated_language=translated_language,
                original_language=original_language,
                excluded_original_language=excluded_original_language,
                content_rating=content_rating,
                excluded_groups=excluded_groups,
                excluded_uploaders=excluded_uploaders,
                include_future_updates=include_future_updates,
                created_at_since=created_at_since,
                updated_at_since=updated_at_since,
                published_at_since=published_at_since,
                order=order,
                includes=includes,
                include_empty_pages=include_empty_pages,
                include_future_publish_at=include_future_publish_at,
                include_external_url=include_external_url,
            )

            from .chapter import Chapter

            chapters.extend([Chapter(self._http, item) for item in data["data"]])

            offset += inner_limit
            if not data["data"] or offset >= 10_000 or limit is not None:
                break

        return ChapterFeed(self._http, data, chapters)

    @require_authentication
    async def update_read_markers(self) -> manga.MangaReadMarkersResponse:
        """|coro|

        This method will return the read chapters of the current manga.

        Returns
        --------
        :class:`hondana.types.manga.MangaReadMarkersResponse`
            The raw payload of the API.
            Contains a list of read chapter UUIDs.
        """
        return await self._http._manga_read_markers([self.id], grouped=False)

    @require_authentication
    async def bulk_update_read_markers(
        self, *, update_history: bool = True, read_chapters: Optional[list[str]], unread_chapters: Optional[list[str]]
    ) -> None:
        """|coro|

        This method will batch update your read chapters for a given Manga.

        Parameters
        -----------
        update_history: :class:`bool`
            Whether to show this chapter in the authenticated user's read history.
            Defaults to ``True``.
        read_chapters: Optional[List[:class:`str`]]
            The read chapters for this Manga.
        unread_chapters: Optional[List[:class:`str`]]
            The unread chapters for this Manga.

        Raises
        -------
        :exc:`TypeError`
            You must provide one or both of the parameters `read_chapters` and/or `unread_chapters`.
        """
        if not read_chapters and not unread_chapters:
            raise TypeError("You must provide either `read_chapters` and/or `unread_chapters` to this method.")

        await self._http._manga_read_markers_batch(
            self.id, update_history=update_history, read_chapters=read_chapters, unread_chapters=unread_chapters
        )

    @require_authentication
    async def get_reading_status(self) -> manga.MangaSingleReadingStatusResponse:
        """|coro|

        This method will return the current reading status for the current manga.

        Raises
        -------
        :exc:`Forbidden`
            You are not authenticated to perform this action.
        :exc:`NotFound`
            The specified manga does not exist, likely due to an incorrect ID.

        Returns
        --------
        :class:`~hondana.types.manga.MangaSingleReadingStatusResponse`
            The raw payload from the API response.
        """
        return await self._http._get_manga_reading_status(self.id)

    @require_authentication
    async def update_reading_status(self, *, status: ReadingStatus) -> None:
        """|coro|

        This method will update your current reading status for the current manga.

        Parameters
        -----------
        status: :class:`~hondana.ReadingStatus`
            The reading status you wish to update this manga with.


        .. note::
            Leaving ``status`` as the default will remove the manga reading status from the API.
            Please provide a value if you do not wish for this to happen.

        Raises
        -------
        :exc:`BadRequest`
            The query parameters were invalid.
        :exc:`NotFound`
            The specified manga cannot be found, likely due to incorrect ID.
        """

        await self._http._update_manga_reading_status(self.id, status=status)

    async def get_volumes_and_chapters(
        self,
        *,
        translated_language: Optional[list[LanguageCode]] = None,
        groups: Optional[list[str]] = None,
    ) -> manga.GetMangaVolumesAndChaptersResponse:
        """|coro|

        This endpoint returns the raw relational mapping of a manga's volumes and chapters.

        Parameters
        -----------
        translated_language: Optional[List[:class:`~hondana.types.common.LanguageCode`]]
            The list of language codes you want to limit the search to.
        groups: Optional[List[:class:`str`]]
            A list of scanlator groups to filter the results by.

        Returns
        --------
        :class:`~hondana.types.manga.GetMangaVolumesAndChaptersResponse`
            The raw payload from mangadex. There is no guarantee of the keys here.
        """
        data = await self._http._get_manga_volumes_and_chapters(
            manga_id=self.id, translated_language=translated_language, groups=groups
        )

        return data

    @require_authentication
    async def add_to_custom_list(self, *, custom_list_id: str) -> None:
        """|coro|

        This method will add the current manga to the specified custom list.

        Parameters
        -----------
        custom_list_id: :class:`str`
            The UUID associated with the custom list you wish to add the manga to.

        Raises
        -------
        :exc:`Forbidden`
            You are not authorised to add manga to this custom list.
        :exc:`NotFound`
            The specified manga or specified custom list are not found, likely due to an incorrect UUID.
        """

        await self._http._add_manga_to_custom_list(manga_id=self.id, custom_list_id=custom_list_id)

    @require_authentication
    async def remove_from_custom_list(self, *, custom_list_id: str) -> None:
        """|coro|

        This method will remove the current manga from the specified custom list.

        Parameters
        -----------
        custom_list_id: :class:`str`
            THe UUID associated with the custom list you wish to add the manga to.

        Raises
        -------
        :exc:`Forbidden`
            You are not authorised to remove a manga from the specified custom list.
        :exc:`NotFound`
            The specified manga or specified custom list are not found, likely due to an incorrect UUID.
        """

        await self._http._remove_manga_from_custom_list(manga_id=self.id, custom_list_id=custom_list_id)

    async def get_chapters(
        self,
        *,
        limit: Optional[int] = 10,
        offset: int = 0,
        ids: Optional[list[str]] = None,
        title: Optional[str] = None,
        groups: Optional[list[str]] = None,
        uploader: Optional[str] = None,
        volumes: Optional[list[str]] = None,
        chapter: Optional[list[str]] = None,
        translated_language: Optional[list[LanguageCode]] = None,
        original_language: Optional[list[LanguageCode]] = None,
        excluded_original_language: Optional[list[LanguageCode]] = None,
        content_rating: Optional[list[ContentRating]] = None,
        excluded_groups: Optional[list[str]] = None,
        excluded_uploaders: Optional[list[str]] = None,
        include_future_updates: Optional[bool] = None,
        created_at_since: Optional[datetime.datetime] = None,
        updated_at_since: Optional[datetime.datetime] = None,
        published_at_since: Optional[datetime.datetime] = None,
        order: Optional[FeedOrderQuery] = None,
        includes: Optional[ChapterIncludes] = ChapterIncludes(),
    ) -> ChapterFeed:
        """|coro|

        This method will return a list of published chapters.

        Parameters
        -----------
        limit: Optional[:class:`int`]
            Defaults to 100. This specifies the amount of chapters to return in one request.
        offset: :class:`int`
            Defaults to 0. This specifies the pagination offset.
        ids: Optional[List[:class:`str`]]
            The list of chapter UUIDs to filter the request with.
        title: Optional[:class:`str`]
            The chapter title query to limit the request with.
        groups: Optional[List[:class:`str`]]
            The scanlation group UUID(s) to limit the request with.
        uploader: Optional[:class:`str`]
            The uploader UUID to limit the request with.
        volumes: Optional[Union[:class:`str`, List[:class:`str`]]]
            The volume UUID or UUIDs to limit the request with.
        chapter: Optional[Union[:class:`str`, List[:class:`str`]]]
            The chapter UUID or UUIDs to limit the request with.
        translated_language: Optional[List[:class:`~hondana.types.common.LanguageCode`]]
            The list of languages codes to filter the request with.
        original_language: Optional[List[:class:`~hondana.types.common.LanguageCode`]]
            The list of languages to specifically target in the request.
        excluded_original_language: Optional[List[:class:`~hondana.types.common.LanguageCode`]]
            The list of original languages to exclude from the request.
        content_rating: Optional[List[:class:`~hondana.ContentRating`]]
            The content rating to filter the feed by.
        excluded_groups: Optional[List[:class:`str`]]
            The list of scanlator groups to exclude from the response.
        excluded_uploaders: Optional[List[:class:`str`]]
            The list of uploaders to exclude from the response.
        include_future_updates: Optional[:class:`bool`]
            Whether to include future chapters in this feed. Defaults to ``True`` API side.
        created_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their creation date.
        updated_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their updated at date.
        published_at_since: Optional[:class:`datetime.datetime`]
            A start point to return chapters from based on their published at date.
        order: Optional[:class:`~hondana.query.FeedOrderQuery`]
            A query parameter to choose how the responses are ordered.
            i.e. ``{"chapters": "desc"}``
        includes: Optional[:class:`~hondana.query.ChapterIncludes`]
            The list of options to include increased payloads for per chapter.
            Defaults to all values.


        .. note::
            Passing ``None`` to ``limit`` will attempt to retrieve all items in the chapter feed.

        .. note::
            If `order` is not specified then the API will return results first based on their creation date,
            which could lead to unexpected results.

        Raises
        -------
        :exc:`BadRequest`
            The query parameters were malformed

        Returns
        --------
        :class:`~hondana.ChapterFeed`
            Returns a collection of chapters.
        """

        inner_limit = limit or 10

        chapters = []
        while True:
            data = await self._http._chapter_list(
                limit=inner_limit,
                offset=offset,
                manga=self.id,
                ids=ids,
                title=title,
                groups=groups,
                uploader=uploader,
                volume=volumes,
                chapter=chapter,
                translated_language=translated_language,
                original_language=original_language,
                excluded_original_language=excluded_original_language,
                content_rating=content_rating,
                excluded_groups=excluded_groups,
                excluded_uploaders=excluded_uploaders,
                include_future_updates=include_future_updates,
                created_at_since=created_at_since,
                updated_at_since=updated_at_since,
                published_at_since=published_at_since,
                order=order,
                includes=includes,
            )
            from .chapter import Chapter

            chapters.extend([Chapter(self._http, item) for item in data["data"]])

            offset += inner_limit
            if not data["data"] or offset >= 10_000 or limit is not None:
                break

        return ChapterFeed(self._http, data, chapters)

    async def get_draft(self) -> Manga:
        """|coro|

        This method will return a manga draft from MangaDex.

        Returns
        --------
        :class:`~hondana.Manga`
            The Manga returned from the API.
        """
        data = await self._http._get_manga_draft(self.id)
        return self.__class__(self._http, data["data"])

    async def submit_draft(self, *, version: int) -> Manga:
        """|coro|

        This method will submit a draft for a manga.

        Parameters
        -----------
        version: :class:`int`
            The version of the manga we're attributing this submission to.

        Raises
        -------
        :exc:`BadRequest`
            The request parameters were incorrect or malformed.
        :exc:`Forbidden`
            You are not authorised to perform this action.
        :exc:`NotFound`
            The manga was not found.

        Returns
        --------
        :class:`~hondana.Manga`
        """
        data = await self._http._submit_manga_draft(self.id, version=version)
        return self.__class__(self._http, data["data"])

    async def get_relations(self, *, includes: Optional[MangaIncludes] = MangaIncludes()) -> MangaRelationCollection:
        """|coro|

        This method will return a list of all relations to a given manga.

        Parameters
        -----------
        includes: Optional[:class:`~hondana.query.MangaIncludes`]
            The optional parameters for expanded requests to the API.
            Defaults to all possible expansions.

        Raises
        -------
        :exc:`BadRequest`
            The manga ID passed is malformed

        Returns
        --------
        :class:`~hondana.MangaRelationCollection`
        """
        data = await self._http._get_manga_relation_list(self.id, includes=includes)
        fmt = [MangaRelation(self._http, self.id, item) for item in data["data"]]
        return MangaRelationCollection(self._http, data, fmt)

    @require_authentication
    async def upload_cover(
        self, *, cover: bytes, volume: Optional[str] = None, description: str, locale: Optional[LanguageCode] = None
    ) -> Cover:
        """|coro|

        This method will upload a cover to the MangaDex API.

        Parameters
        -----------
        cover: :class:`bytes`
            THe raw bytes of the image.
        volume: Optional[:class:`str`]
            The volume this cover relates to.
        description: :class:`str`
            The description of this cover.
        locale: Optional[:class:`~hondana.types.common.LanguageCode`]
            The locale of this cover.

        Raises
        -------
        :exc:`BadRequest`
            The volume parameter was malformed or the file was a bad format.
        :exc:`Forbidden`
            You are not permitted for this action.

        Returns
        --------
        :class:`~hondana.Cover`
        """
        data = await self._http._upload_cover(self.id, cover=cover, volume=volume, description=description, locale=locale)
        return Cover(self._http, data["data"])

    @require_authentication
    async def create_relation(self, *, target_manga: str, relation_type: MangaRelationType) -> MangaRelation:
        """|coro|

        This method will create a manga relation.

        Parameters
        ------------
        target_manga: :class:`str`
            The manga ID of the related manga.
        relation_type: :class:`~hondana.MangaRelationType`

        Raises
        -------
        :exc:`BadRequest`
            The parameters were malformed
        :exc:`Forbidden`
            You are not authorised for this action.

        Returns
        --------
        :class:`~hondana.MangaRelation`
        """
        data = await self._http._create_manga_relation(self.id, target_manga=target_manga, relation_type=relation_type)
        return MangaRelation(self._http, self.id, data["data"])

    @require_authentication
    async def delete_relation(self, relation_id: str, /) -> None:
        """|coro|

        This method will delete a manga relation.

        Parameters
        -----------
        relation_id: :class:`str`
            The ID of the related manga.
        """
        await self._http._delete_manga_relation(self.id, relation_id)

    @require_authentication
    async def set_rating(self, *, rating: int) -> None:
        """|coro|

        This method will set your rating on the manga.
        This method **overwrites** your previous set rating, if any.

        Parameters
        -----------
        rating: :class:`int`
            The rating value, between 0 and 10.

        Raises
        -------
        :exc:`Forbidden`
            The request returned a response due to authentication failure.
        :exc:`NotFound`
            The specified manga UUID was not found or does not exist.
        """
        await self._http._set_manga_rating(self.id, rating=rating)

    @require_authentication
    async def delete_rating(self) -> None:
        """|coro|

        This method will delete your set rating on the manga.

        Raises
        -------
        :exc:`Forbidden`
            The request returned a response due to authentication failure.
        :exc:`NotFound`
            The specified manga UUID was not found or does not exist.
        """
        await self._http._delete_manga_rating(self.id)

    async def get_statistics(self) -> MangaStatistics:
        """|coro|

        This method will fetch statistics on the current manga, and cache them as the :attr:`stats`

        Returns
        --------
        :class:`~hondana.MangaStatistics`
        """
        data = await self._http._get_manga_statistics(self.id)

        key = next(iter(data["statistics"]))
        stats = MangaStatistics(self._http, self.id, data["statistics"][key])

        self.stats = stats
        return self.stats


class MangaRelation:
    """A class representing a MangaRelation returned from the MangaDex API.

    Attributes
    -----------
    source_manga_id: :class:`str`
        The UUID associated to the parent manga of this relation.
    id: :class:`str`
        The UUID associated to this manga relation.
    version: :class:`int`
        The version revision of this manga relation.
    relation_type: :class:`~hondana.MangaRelationType`
        The type of relationship to the source manga.
    """

    __slots__ = (
        "_http",
        "_data",
        "_attributes",
        "_relationships",
        "source_manga_id",
        "id",
        "version",
        "relation_type",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: manga.MangaRelation, /) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        self._relationships = self._data.pop("relationships", [])
        self.source_manga_id: str = parent_id
        self.id: str = self._data["id"]
        self.version: int = self._attributes["version"]
        self.relation_type: MangaRelationType = MangaRelationType(self._attributes["relation"])

    def __repr__(self) -> str:
        return f"<MangaRelation id={self.id!r} source_manga_id={self.source_manga_id!r}>"

    def __eq__(self, other: Union[MangaRelation, Manga]) -> bool:
        return self.id == other.id or self.source_manga_id == other.id

    def __ne__(self, other: Union[MangaRelation, Manga]) -> bool:
        return not self.__eq__(other)


class MangaStatistics:
    """
    A small object to house manga statistics.

    Attributes
    -----------
    follows: :class:`int`
        The number of follows this manga has.
    parent_id: :class:`str`
        The manga these statistics belong to.
    average: Optional[:class:`float`]
        The average mean score of the manga ratings.
    bayesian: Optional[:class:`float`]
        The bayesian average score of the manga ratings.
    distribution: Optional[Dict[:class:`str`, :class:`int`]]
        The scoring distribution of the manga.
        Keys are 1-10 and values are total amount of ratings per key.


    .. note::
        The :attr:`distribution` attribute will be None unless this object was created with :meth:`~hondana.Client.get_manga_statistics` or :meth:`~hondana.Manga.get_statistics`
    """

    __slots__ = (
        "_http",
        "_data",
        "_rating",
        "follows",
        "parent_id",
        "average",
        "bayesian",
        "distribution",
    )

    def __init__(
        self, http: HTTPClient, parent_id: str, payload: Union[StatisticsResponse, BatchStatisticsResponse]
    ) -> None:
        self._http = http
        self._data = payload
        self._rating = payload["rating"]
        self.follows: int = payload["follows"]
        self.parent_id: str = parent_id
        self.average: Optional[float] = self._rating["average"]
        self.bayesian: Optional[float] = self._rating["bayesian"]
        self.distribution: Optional[dict[str, int]] = self._rating.get("distribution")

    def __repr__(self) -> str:
        return f"<MangaStatistics for={self.parent_id!r}>"


class MangaRating:
    """
    A small object to encompass your personal manga ratings.

    Attributes
    -----------
    parent_id: :class:`str`
        The parent manga this rating belongs to.
    rating: :class:`int`
        Your personal rating for this manga, between 1 and 10.
    created_at: :class:`datetime.datetime`
        When you created the rating, as a UTC aware datetime.
    """

    __slots__ = (
        "_http",
        "_data",
        "parent_id",
        "rating",
        "created_at",
    )

    def __init__(self, http: HTTPClient, parent_id: str, payload: PersonalMangaRatingsResponse) -> None:
        self._http = http
        self._data = payload
        self.parent_id: str = parent_id
        self.rating: int = self._data["rating"]
        self.created_at: datetime.datetime = datetime.datetime.fromisoformat(self._data["createdAt"])

    def __repr__(self) -> str:
        return f"<MangaRating parent_id={self.parent_id!r}>"



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from . import account as account
from . import auth as auth
from . import author as author
from . import chapter as chapter
from . import common as common
from . import cover as cover
from . import custom_list as custom_list
from . import errors as errors
from . import legacy as legacy
from . import manga as manga
from . import relationship as relationship
from . import report as report
from . import scanlator_group as scanlator_group
from . import settings as settings
from . import statistics as statistics
from . import tags as tags
from . import token as token
from . import upload as upload
from . import user as user



"""
The MIT License (MIT)

Copyright (c) 2021-Present AbstractUmbra

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Optional

from .user import User
from .utils import MISSING, relationship_finder, require_authentication


if TYPE_CHECKING:
    from .http import HTTPClient
    from .types.common import LanguageCode
    from .types.cover import CoverResponse
    from .types.manga import MangaResponse
    from .types.user import UserResponse


__all__ = ("Cover",)


class Cover:
    """A class representing a Cover returned from the MangaDex API.

    Attributes
    -----------
    id: :class:`str`
        The UUID associated with this cover.
    volume: Optional[:class:`str`]
        The volume attributed to this cover, if any.
    file_name: :class:`str`
        The file name of this cover.
    description: :class:`str`
        The description of this cover.
    version: :class:`int`
        The version revision of this Cover.
    """

    __slots__ = (
        "_http",
        "_data",
        "_attributes",
        "id",
        "volume",
        "file_name",
        "description",
        "locale",
        "version",
        "_created_at",
        "_updated_at",
        "_manga_relationship",
        "_uploader_relationship",
    )

    def __init__(self, http: HTTPClient, payload: CoverResponse) -> None:
        self._http = http
        self._data = payload
        self._attributes = self._data["attributes"]
        relationships = self._data.pop("relationships", [])
        self.id: str = self._data["id"]
        self.volume: Optional[str] = self._attributes["volume"]
        self.file_name: str = self._attributes["fileName"]
        self.description: Optional[str] = self._attributes["description"]
        self.locale: Optional[LanguageCode] = self._attributes["locale"]
        self.version: int = self._attributes["version"]
        self._created_at = self._attributes["createdAt"]
        self._updated_at = self._attributes["updatedAt"]
        self._manga_relationship: Optional[MangaResponse] = relationship_finder(relationships, "manga", limit=1)
        self._uploader_relationship: Optional[UserResponse] = relationship_finder(relationships, "user", limit=1)

    def __repr__(self) -> str:
        return f"<Cover id={self.id!r} filename={self.file_name!r}>"

    def __str__(self) -> str:
        return self.file_name

    def __eq__(self, other: Cover) -> bool:
        return isinstance(other, Cover) and self.id == other.id

    def __ne__(self, other: Cover) -> bool:
        return not self.__eq__(other)

    @property
    def created_at(self) -> datetime.datetime:
        """When this cover was created.

        Returns
        --------
        :class:`datetime.datetime`
            The UTC datetime of when this cover was created.
        """
        return datetime.datetime.fromisoformat(self._created_at)

    @property
    def updated_at(self) -> datetime.datetime:
        """When this cover was last updated.

        Returns
        --------
        :class:`datetime.datetime`
            The UTC datetime of when this cover was last updated.
        """
        return datetime.datetime.fromisoformat(self._updated_at)

    @property
    def uploader(self) -> Optional[User]:
        """The user who uploaded this cover.

        .. note::
            This is only populated if the Cover has populated relationships.

        Returns
        --------
        Optional[:class:`~hondana.User`]
            The user who uploaded this cover, if present.
        """
        if not self._uploader_relationship:
            return

        if "attributes" in self._uploader_relationship:
            return User(self._http, self._uploader_relationship)

    def url(self, type: Optional[Literal[256, 512]] = None, /, parent_id: Optional[str] = None) -> Optional[str]:
        """Method to return the Cover url.

        Due to the API structure, this will return ``None`` if the parent manga key is missing from the response relationships.

        Parameters
        -----------
        type: Optional[Literal[``256``, ``512``]]
            Defaults to ``None`` to return original quality.
            Specifies the return image dimensions.
        parent_id: Optional[:class:`str`]
            Parameter where you can potentially set or override the parent manga to skip relationship lookup.
            Useful for when Cover is part of something else's relationships, holding none of it's own.

        Returns
        --------
        Optional[:class:`str`]
            The Cover url.
        """
        manga_id = parent_id or (self._manga_relationship["id"] if self._manga_relationship else None)
        if manga_id is None:
            return

        if type == 256:
            fmt = ".256.jpg"
        elif type == 512:
            fmt = ".512.jpg"
        else:
            fmt = ""

        return f"https://uploads.mangadex.org/covers/{manga_id}/{self.file_name}{fmt}"

    async def fetch_image(self, size: Optional[Literal[256, 512]] = None, /) -> Optional[bytes]:
        """|coro|

        This method depends on :attr:`url`, as such, it can return None.

        Parameters
        -----------
        size: Optional[Literal[``256``, ``512``]]
            Defaults to ``None`` to return original quality.
            Specifies the returned image size.

        Returns
        --------
        :class:`bytes`
            The raw image bytes.
        """
        url = self.url(size)
        if url is None:
            return

        assert self._http._session is not None  # if you got this far, this cannot be None
        async with self._http._session.get(url) as resp:
            data = await resp.read()

        return data

    @require_authentication
    async def edit_cover(
        self, *, volume: Optional[str] = MISSING, description: Optional[str] = MISSING, version: int
    ) -> Cover:
        """|coro|

        This method will edit the current cover on the MangaDex API.

        Parameters
        -----------
        volume: :class:`str`
            The volume identifier relating the cover will represent.
        description: Optional[:class:`str`]
            The description of the cover.
        version: :class:`int`
            The version revision of the cover.


        .. warning::
            The ``volume`` key is mandatory. You can pass ``None`` to null it in the API, but it must have a value.

        Raises
        -------
        TypeError
            The volume key was not given a value. This is required.
        :exc:`BadRequest`
            The request body was malformed.
        :exc:`Forbidden`
            The request returned an error due to authentication failure.

        Returns
        --------
        :class:`~hondana.Cover`
            The returned cover after the edit.
        """
        data = await self._http._edit_cover(self.id, volume=volume, description=description, version=version)

        return self.__class__(self._http, data["data"])

    @require_authentication
    async def delete(self) -> None:
        """|coro|

        This method will delete the current cover from the MangaDex API.

        Raises
        -------
        :exc:`BadRequest`
            The request payload was malformed.
        :exc:`Forbidden`
            The request returned an error due to authentication.
        """
        await self._http._delete_cover(self.id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hondana', 'hondana.types']

package_data = \
{'': ['*'], 'hondana': ['extras/*']}

modules = \
['py', 'tags', 'report_reasons']
install_requires = \
['aiohttp>=3.8.1,<4.0.0', 'typing-extensions']

entry_points = \
{'console_scripts': ['version = hondana.__main__:show_version']}

setup_kwargs = {
    'name': 'hondana',
    'version': '3.4.3',
    'description': 'An asynchronous wrapper around the MangaDex v5 API',
    'long_description': '<div align="center">\n    <h1><a href="https://jisho.org/word/%E6%9C%AC%E6%A3%9A">Hondana 『本棚』</a></h1>\n    <a href=\'https://github.com/AbstractUmbra/Hondana/actions/workflows/build.yaml\'>\n        <img src=\'https://github.com/AbstractUmbra/Hondana/actions/workflows/build.yaml/badge.svg\' alt=\'Build status\' />\n    </a>\n    <a href=\'https://github.com/AbstractUmbra/Hondana/actions/workflows/coverage_and_lint.yaml\'>\n        <img src=\'https://github.com/AbstractUmbra/Hondana/actions/workflows/coverage_and_lint.yaml/badge.svg\' alt=\'Linting and Typechecking\' />\n    </a>\n</div>\n<div align="center">\n    <a href=\'https://api.mangadex.org/\'>\n        <img src=\'https://img.shields.io/website?down_color=red&down_message=offline&label=API%20Status&logo=MangaDex%20API&up_color=lime&up_message=online&url=https%3A%2F%2Fapi.mangadex.org%2Fping\' alt=\'API Status\'/>\n    </a>\n    <a href=\'https://hondana.readthedocs.io/en/latest/?badge=latest\'>\n        <img src=\'https://readthedocs.org/projects/hondana/badge/?version=latest\' alt=\'Documentation Status\' />\n    </a>\n</div>\n<h1></h1>\n<br>\n\nA lightweight and asynchronous wrapper around the [MangaDex v5 API](https://api.mangadex.org/docs.html).\nYou can see our stable docs [here](https://hondana.readthedocs.io/en/stable/)!\n\n## Features\nWe are currently at 100% feature compliance with the API.\n\n## Notes\n### Authenticated endpoints\nSadly (thankfully?) I am not an author on MangaDex, meaning I cannot test the creation endpoints for things like scanlators, artists, authors, manga or chapters.\nI have followed the API guidelines to the letter for these, but they may not work.\n\nAny help in testing them is greatly appreciated.\n\n### Upload & Creation\nFollowing the above, I can only use the public test manga for upload testing.\nThese are currently implemented and tested to the best of my ability.\n\n## Examples\nPlease take a look at the [examples](./examples/) directory for working examples.\n\n**NOTE**: More examples will follow as the library is developed.\n\n### Current caveats to note\n- There are no API endpoints for Artist. It seems they are not differentiated from Author types except in name only.\n  - I have separated them logically, but under the hood all Artists are Authors and their `__eq__` reports as such.\n- The tags and report reasons are locally cached since you **must** pass UUIDs to the api (and I do not think you\'re going to memorize those), there\'s a convenience method for updating the local cache as `Client.update_tags` and `Client.update_report_reasons` respectively.\n  - I have added [an example](./examples/updating_local_tags.py) on how to do the above for tags.\n  - To use these tags, you can see an example [here](./examples/search_manga.py#L17-L22).\n- [Settings related endpoints](https://api.mangadex.org/docs.html#operation/get-settings-template) are not currently exposed. I have implemented their documented use, but I do not expect them to currently function.\n  - Once this is exposed fully I will implement a richer interface.\n- Currently, if there are any errors in a chapter upload process, the error key does not contain filenames or identifiable information on exactly what file failed upload.\n  - This means that I locally compare the succeeded files with the ones missing from the full response payload. The examples have been updated with how to check for an error.\n- Currently, the `Client.my_chapter_read_history` method will not work. It is disabled on MD\'s side due to an issue they had previously. If/when it is re-introduced the method will remain the same.\n\n### Further information/tidbits\n- For a bit more clarity on a Chapter\'s `readableAt` vs `publishAt`, see [this page](https://api.mangadex.org/docs/dates/#readableat) on the MangaDex docs.\n- A query with the `include_future_updates` bool set to `True` will include chapters that are pending release by scanlation groups on MangaDex, but which may be available on their sites.\n\n### Contributing\nIf you would like to contribute to Hondana, please take a look at [the contributing guidelines](./.github/CONTRIBUTING.md) and follow the procedure there.\n\n\nIf you have any question please feel free to join my Discord server:\n<div align="left">\n    <a href="https://discord.gg/aYGYJxwqe5">\n        <img src="https://discordapp.com/api/guilds/705500489248145459/widget.png?style=banner2" alt="Discord Server"/>\n    </a>\n</div>\n',
    'author': 'Alex Nørgaard',
    'author_email': 'Umbra@AbstractUmbra.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/AbstractUmbra/hondana',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
